# -*- coding: utf-8 -*-
"""This module contains an abstraction class Linop for linear operators,
and provides commonly used linear operators, including signal transforms
such as FFT, NUFFT, and wavelet, and array manipulation operators,
such as reshape, transpose, and resize.
"""
import numpy as np

from sigpy import backend, block, conv, fourier, interp, util, wavelet


def _check_shape_positive(shape):
    if not all(s > 0 for s in shape):
        raise ValueError(
            "Shapes must be positive, got {shape}".format(shape=shape)
        )


class Linop:
    """Abstraction for linear operator.

    Linop can be called on a NumPy or CuPy array to
    perform a linear operation. The output can be on a different device.

    Given a Linop A, and an appropriately shaped input x, the following are
    both valid operations to compute x -> A(x):

       >>> y = A * x
       >>> y = A(x)

    Its adjoint linear operator can be obtained using the .H attribute.
    Linops can be scaled, added, subtracted, stacked and composed.
    Here are some example of valid operations on Linop A, Linop B,
    and a scalar a:

       >>> A.H
       >>> a * A + B
       >>> a * A * B

    Args:
        oshape: Output shape.
        ishape: Input shape.
        repr_str (string or None): default: class name.

    Attributes:
        oshape: output shape.
        ishape: input shape.
        H: adjoint linear operator.
        N: normal linear operator.

    """

    def __init__(self, oshape, ishape, repr_str=None):
        self.oshape = list(oshape)
        self.ishape = list(ishape)

        _check_shape_positive(oshape)
        _check_shape_positive(ishape)

        if repr_str is None:
            self.repr_str = self.__class__.__name__
        else:
            self.repr_str = repr_str

        self.adj = None
        self.normal = None

    def _check_ishape(self, input):
        for i1, i2 in zip(input.shape, self.ishape):
            if i2 != -1 and i1 != i2:
                raise ValueError(
                    "input shape mismatch for {s}, got {input_shape}".format(
                        s=self, input_shape=input.shape
                    )
                )

    def _check_oshape(self, output):
        for o1, o2 in zip(output.shape, self.oshape):
            if o2 != -1 and o1 != o2:
                raise ValueError(
                    "output shape mismatch for {s}, got {output_shape}".format(
                        s=self, output_shape=output.shape
                    )
                )

    def _apply(self, input):
        raise NotImplementedError

    def apply(self, input):
        """Apply linear operation on input.

        This function checks for the input/output shapes and devices,
        and calls the internal user-defined _apply() method.

        Args:
            input (array): input array of shape `ishape`.

        Returns:
            array: output array of shape `oshape`.

        """
        try:
            self._check_ishape(input)
            output = self._apply(input)
            self._check_oshape(output)
        except Exception as e:
            raise RuntimeError("Exceptions from {}.".format(self)) from e

        return output

    def _adjoint_linop(self):
        raise NotImplementedError

    def _normal_linop(self):
        return self.H * self

    @property
    def H(self):
        r"""Return adjoint linear operator.

        An adjoint linear operator :math:`A^H` for
        a linear operator :math:`A` is defined as:

        .. math:
            \left< A x, y \right> = \left< x, A^H, y \right>

        Returns:
            Linop: adjoint linear operator.

        """
        if self.adj is None:
            self.adj = self._adjoint_linop()
        return self.adj

    @property
    def N(self):
        r"""Return normal linear operator.

        A normal linear operator :math:`A^HA` for
        a linear operator :math:`A`.

        Returns:
            Linop: adjoint linear operator.

        """
        if self.normal is None:
            self.normal = self._normal_linop()
        return self.normal

    def __call__(self, input):
        return self.__mul__(input)

    def __mul__(self, input):
        if isinstance(input, Linop):
            return Compose([self, input])
        elif np.isscalar(input):
            M = Multiply(self.ishape, input)
            return Compose([self, M])
        elif isinstance(input, backend.get_array_module(input).ndarray):
            return self.apply(input)

        return NotImplemented

    def __rmul__(self, input):
        if np.isscalar(input):
            M = Multiply(self.oshape, input)
            return Compose([M, self])

        return NotImplemented

    def __add__(self, input):
        if isinstance(input, Linop):
            return Add([self, input])
        else:
            raise NotImplementedError

    def __neg__(self):
        return -1 * self

    def __sub__(self, input):
        return self.__add__(-input)

    def __repr__(self):
        return "<{oshape}x{ishape}> {repr_str} Linop>".format(
            oshape=self.oshape, ishape=self.ishape, repr_str=self.repr_str
        )


class Identity(Linop):
    """Identity linear operator.

    Returns input directly.

    Args:
        shape (tuple of ints): Input shape

    """

    def __init__(self, shape):
        super().__init__(shape, shape)

    def _apply(self, input):
        return input

    def _adjoint_linop(self):
        return self

    def _normal_linop(self):
        return self


class ToDevice(Linop):
    """Move input between devices.

    Args:
        shape (tuple of ints): Input/output shape.
        odevice (Device): Output device
        idevice (Device): Input device
    """

    def __init__(self, shape, odevice, idevice):
        self.odevice = odevice
        self.idevice = idevice

        super().__init__(shape, shape)

    def _apply(self, input):
        return backend.to_device(input, self.odevice)

    def _adjoint_linop(self):
        return ToDevice(self.ishape, self.idevice, self.odevice)


class AllReduce(Linop):
    """Performs all reduce between MPI ranks.

    Returns input if in_place is True.
    Otherwise returns reduced array on the input device.

    Args:
        shape (tuple of ints): Input/output shape.
        comm (Communicator): Communicator.

    """

    def __init__(self, shape, comm, in_place=False):
        self.comm = comm
        self.in_place = in_place

        super().__init__(shape, shape)

    def _apply(self, input):
        with backend.get_device(input):
            if self.in_place:
                output = input
            else:
                output = input.copy()

            self.comm.allreduce(output)
            return output

    def _adjoint_linop(self):
        return AllReduceAdjoint(self.ishape, self.comm, in_place=self.in_place)


class AllReduceAdjoint(Linop):
    """All reduce adjoint operator. Equivalant to identity.

    Args:
        shape (tuple of ints): Input/output shape.
        comm (Communicator): Communicator.

    """

    def __init__(self, shape, comm, in_place=False):
        self.comm = comm
        self.in_place = in_place

        super().__init__(shape, shape)

    def _apply(self, input):
        return input

    def _adjoint_linop(self):
        return AllReduce(self.ishape, self.comm, in_place=self.in_place)


class Conj(Linop):
    """Complex conjugate of linear operator.

    Args:
        A (Linop): Input linear operator.

    """

    def __init__(self, A):
        self.A = A

        super().__init__(A.oshape, A.ishape, repr_str=A.repr_str)

    def _apply(self, input):
        device = backend.get_device(input)
        xp = device.xp
        with device:
            input = xp.conj(input)

        output = self.A(input)

        device = backend.get_device(output)
        xp = device.xp
        with device:
            return xp.conj(output)

    def _adjoint_linop(self):
        return Conj(self.A.H)


class Add(Linop):
    """Addition of linear operators.

    Input and output shapes and devices must match.

    Args:
        linops (list of Linops): Input linear operators.

    Returns:
        Linop: linops[0] + linops[1] + ... + linops[n - 1]

    """

    def __init__(self, linops):
        _check_linops_same_ishape(linops)
        _check_linops_same_oshape(linops)

        self.linops = linops
        oshape = linops[0].oshape
        ishape = linops[0].ishape

        super().__init__(
            oshape,
            ishape,
            repr_str=" + ".join([linop.repr_str for linop in linops]),
        )

    def _apply(self, input):
        output = 0
        with backend.get_device(input):
            for linop in self.linops:
                output += linop(input)

        return output

    def _adjoint_linop(self):
        return Add([linop.H for linop in self.linops])


def _check_compose_linops(linops):
    for linop1, linop2 in zip(linops[:-1], linops[1:]):
        if linop1.ishape != linop2.oshape:
            raise ValueError(
                "cannot compose {linop1} and {linop2}.".format(
                    linop1=linop1, linop2=linop2
                )
            )


def _combine_compose_linops(linops):
    combined_linops = []
    for linop in linops:
        if isinstance(linop, Compose):
            combined_linops += linop.linops
        else:
            combined_linops.append(linop)

    return combined_linops


class Compose(Linop):
    """Composition of linear operators.

    Args:
        linops (list of Linops): Linear operators to be composed.

    Returns:
        Linop: linops[0] * linops[1] * ... * linops[n - 1]

    """

    def __init__(self, linops):
        _check_compose_linops(linops)
        self.linops = _combine_compose_linops(linops)

        super().__init__(
            self.linops[0].oshape,
            self.linops[-1].ishape,
            repr_str=" * ".join([linop.repr_str for linop in linops]),
        )

    def _apply(self, input):
        output = input
        for linop in self.linops[::-1]:
            output = linop(output)

        return output

    def _adjoint_linop(self):
        return Compose([linop.H for linop in self.linops[::-1]])


def _check_linops_same_ishape(linops):
    for linop in linops:
        if linop.ishape != linops[0].ishape:
            raise ValueError(
                "Linops must have the same ishape, got {linops}.".format(
                    linops=linops
                )
            )


def _check_linops_same_oshape(linops):
    for linop in linops:
        if linop.oshape != linops[0].oshape:
            raise ValueError(
                "Linops must have the same oshape, got {linops}.".format(
                    linops=linops
                )
            )


def _hstack_params(shapes, axis):
    if axis is None:
        return _hstack_params([[util.prod(shape)] for shape in shapes], 0)

    ishape = list(shapes[0])
    ndim = len(ishape)
    idx = shapes[0][axis]
    indices = []

    for shape in shapes[1:]:
        if len(shape) != ndim:
            raise Exception(
                "Shapes must have the same lengths to concatenate."
            )

        for i in range(ndim):
            if i == axis:
                ishape[i] += shape[i]
                indices.append(idx)
                idx += shape[i]
            elif shape[i] != ishape[i]:
                raise RuntimeError(
                    "Shapes not along axis must be the same to concatenate."
                )

    return ishape, indices


class Hstack(Linop):
    """Horizontally stack linear operators.

    Creates a Linop that splits the input, applies Linops independently,
    and sums outputs.
    In matrix form, this is equivalant to given matrices {A1, ..., An},
    returns [A1, ..., An].

    Input and output devices must be the same.

    Args:
        linops (list of Linops): list of linops with the same output shape.
        axis (int or None): If None, inputs are vectorized and concatenated.
            Otherwise, inputs are stacked along axis.

    """

    def __init__(self, linops, axis=None):
        self.nops = len(linops)
        _check_linops_same_oshape(linops)

        self.linops = linops
        self.axis = axis

        ishape, self.indices = _hstack_params(
            [linop.ishape for linop in self.linops], axis
        )
        oshape = self.linops[0].oshape

        super().__init__(oshape, ishape)

    def _apply(self, input):
        with backend.get_device(input):
            output = 0
            for n, linop in enumerate(self.linops):
                if n == 0:
                    start = 0
                else:
                    start = self.indices[n - 1]

                if n == self.nops - 1:
                    end = None
                else:
                    end = self.indices[n]

                if self.axis is None:
                    output += linop(input[start:end].reshape(linop.ishape))
                else:
                    ndim = len(linop.ishape)
                    axis = self.axis % ndim

                    slc = tuple(
                        [slice(None)] * axis
                        + [slice(start, end)]
                        + [slice(None)] * (ndim - axis - 1)
                    )

                    output += linop(input[slc])

            return output

    def _adjoint_linop(self):
        return Vstack([op.H for op in self.linops], axis=self.axis)


def _vstack_params(shapes, axis):
    if axis is None:
        return _vstack_params([[util.prod(shape)] for shape in shapes], 0)

    oshape = list(shapes[0])
    ndim = len(oshape)
    idx = shapes[0][axis]
    indices = []

    for shape in shapes[1:]:
        if len(shape) != ndim:
            raise Exception(
                "Shapes must have the same lengths to concatenate."
            )

        for i in range(ndim):
            if i == axis:
                oshape[i] += shape[i]
                indices.append(idx)
                idx += shape[i]
            elif shape[i] != oshape[i]:
                raise Exception(
                    "Shapes not along axis must be the same to concatenate."
                )

    return oshape, indices


class Vstack(Linop):
    """Vertically stack linear operators.

    Creates a Linop that applies linops independently,
    and concatenates outputs.
    In matrix form, this is equivalant to given matrices {A1, ..., An},
    returns [A1.T, ..., An.T].T.

    Args:
        linops (list of Linops): list of linops with the same input shape.
        axis (int or None): If None, outputs are vectorized and concatenated.

    """

    def __init__(self, linops, axis=None):
        self.nops = len(linops)
        _check_linops_same_ishape(linops)

        self.axis = axis
        self.linops = linops

        oshape, self.indices = _vstack_params(
            [linop.oshape for linop in self.linops], axis
        )
        ishape = self.linops[0].ishape

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        xp = device.xp
        with device:
            output = xp.empty(self.oshape, dtype=input.dtype)
            for n, linop in enumerate(self.linops):
                if n == 0:
                    start = 0
                else:
                    start = self.indices[n - 1]

                if n == self.nops - 1:
                    end = None
                else:
                    end = self.indices[n]

                if self.axis is None:
                    output[start:end] = linop(input).ravel()
                else:
                    ndim = len(linop.oshape)
                    axis = self.axis % ndim
                    slc = tuple(
                        [slice(None)] * axis
                        + [slice(start, end)]
                        + [slice(None)] * (ndim - axis - 1)
                    )
                    output[slc] = linop(input)

        return output

    def _adjoint_linop(self):
        return Hstack([op.H for op in self.linops], axis=self.axis)


class Diag(Linop):
    """Diagonally stack linear operators.

    Create a Linop that splits input, applies linops independently,
    and concatenates outputs.
    In matrix form, given matrices {A1, ..., An}, returns diag([A1, ..., An]).

    Args:
        linops (list of Linops): list of linops with the same input and
            output shape.
        iaxis (int or None): If None, inputs are vectorized
            and concatenated.
        oaxis (int or None): If None, outputs are vectorized
            and concatenated.

    """

    def __init__(self, linops, oaxis=None, iaxis=None):
        self.nops = len(linops)

        self.linops = linops
        self.oaxis = oaxis
        self.iaxis = iaxis
        ishape, self.iindices = _hstack_params(
            [linop.ishape for linop in self.linops], iaxis
        )
        oshape, self.oindices = _vstack_params(
            [linop.oshape for linop in self.linops], oaxis
        )

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        xp = device.xp
        with device:
            output = xp.empty(self.oshape, dtype=input.dtype)
            for n, linop in enumerate(self.linops):
                if n == 0:
                    istart = 0
                    ostart = 0
                else:
                    istart = self.iindices[n - 1]
                    ostart = self.oindices[n - 1]

                if n == self.nops - 1:
                    iend = None
                    oend = None
                else:
                    iend = self.iindices[n]
                    oend = self.oindices[n]

                if self.iaxis is None:
                    output_n = linop(
                        input[istart:iend].reshape(linop.ishape)
                    ).ravel()
                else:
                    ndim = len(linop.ishape)
                    axis = self.iaxis % ndim
                    islc = tuple(
                        [slice(None)] * axis
                        + [slice(istart, iend)]
                        + [slice(None)] * (ndim - axis - 1)
                    )

                    output_n = linop(input[islc])

                if self.oaxis is None:
                    output[ostart:oend] = output_n
                else:
                    ndim = len(linop.oshape)
                    axis = self.oaxis % ndim
                    oslc = tuple(
                        [slice(None)] * axis
                        + [slice(ostart, oend)]
                        + [slice(None)] * (ndim - axis - 1)
                    )

                    output[oslc] = output_n

            return output

    def _adjoint_linop(self):
        return Diag(
            [op.H for op in self.linops], oaxis=self.iaxis, iaxis=self.oaxis
        )


class Reshape(Linop):
    """Reshape input to given output shape.

    Args:
        oshape (tuple of ints): Output shape.
        ishape (tuple of ints): Input shape.

    """

    def __init__(self, oshape, ishape):
        super().__init__(oshape, ishape)

    def _apply(self, input):
        with backend.get_device(input):
            return input.reshape(self.oshape)

    def _adjoint_linop(self):
        return Reshape(self.ishape, self.oshape)

    def _normal_linop(self):
        return Identity(self.ishape)


class Transpose(Linop):
    """Tranpose input with the given axes.

    Args:
        ishape (tuple of ints): Input shape.
        axes (None or tuple of ints): Axes to transpose input.

    """

    def __init__(self, ishape, axes=None):
        self.axes = axes
        if axes is None:
            self.iaxes = None
            oshape = ishape[::-1]
        else:
            self.iaxes = np.argsort(axes)
            oshape = [ishape[a] for a in axes]

        super().__init__(oshape, ishape)

    def _apply(self, input):
        return input.transpose(self.axes)

    def _adjoint_linop(self):
        if self.axes is None:
            iaxes = None
            oshape = self.ishape[::-1]
        else:
            iaxes = np.argsort(self.axes)
            oshape = [self.ishape[a] for a in self.axes]

        return Transpose(oshape, axes=iaxes)

    def _normal_linop(self):
        return Identity(self.ishape)


class FFT(Linop):
    """FFT linear operator.

    Args:
        ishape (tuple of int): Input shape
        axes (None or tuple of int): Axes to perform FFT.
            If None, applies on all axes.
        center (bool): Toggle center FFT.

    """

    def __init__(self, shape, axes=None, center=True):
        self.axes = axes
        self.center = center

        super().__init__(shape, shape)

    def _apply(self, input):
        with backend.get_device(input):
            return fourier.fft(input, axes=self.axes, center=self.center)

    def _adjoint_linop(self):
        return IFFT(self.ishape, axes=self.axes, center=self.center)

    def _normal_linop(self):
        return Identity(self.ishape)


class IFFT(Linop):
    """IFFT linear operator.

    Args:
        ishape (tuple of int): Input shape
        axes (None or tuple of int): Axes to perform FFT.
            If None, applies on all axes.
        center (bool): Toggle center FFT.

    """

    def __init__(self, shape, axes=None, center=True):
        self.axes = axes
        self.center = center

        super().__init__(shape, shape)

    def _apply(self, input):
        with backend.get_device(input):
            return fourier.ifft(input, axes=self.axes, center=self.center)

    def _adjoint_linop(self):
        return FFT(self.ishape, axes=self.axes, center=self.center)

    def _normal_linop(self):
        return Identity(self.ishape)


def _get_matmul_oshape(ishape, mshape, adjoint):
    ishape_exp, mshape_exp = util._expand_shapes(ishape, mshape)
    if adjoint:
        mshape_exp[-1], mshape_exp[-2] = mshape_exp[-2], mshape_exp[-1]

    oshape = []
    for i, m in zip(ishape_exp[:-2], mshape_exp[:-2]):
        if not (i == m or i == 1 or m == 1):
            raise ValueError(
                "Invalid shapes: {ishape}, {mshape}.".format(
                    ishape=ishape, mshape=mshape
                )
            )

        oshape.append(max(i, m))

    if mshape_exp[-1] != ishape_exp[-2]:
        raise ValueError(
            "Invalid shapes: {ishape}, {mshape}.".format(
                ishape=ishape, mshape=mshape
            )
        )

    oshape += [mshape_exp[-2], ishape_exp[-1]]

    return oshape


def _get_matmul_adjoint_sum_axes(oshape, ishape, mshape):
    ishape_exp, mshape_exp = util._expand_shapes(ishape, mshape)
    max_ndim = max(len(ishape), len(mshape))
    sum_axes = []
    for i, m, o, d in zip(
        ishape_exp[:-2], mshape_exp[:-2], oshape[:-2], range(max_ndim - 2)
    ):
        if i == 1 and (m != 1 or o != 1):
            sum_axes.append(d)

    return sum_axes


class MatMul(Linop):
    """Matrix multiplication.

    Args:
        ishape (tuple of ints): Input shape.
            It must be able to broadcast with mat.shape.
        mat (array): Matrix of shape [..., m, n]
        adjoint (bool): Toggle adjoint.
            If True, performs conj(mat).swapaxes(-1, -2)
            before performing matrix multiplication.

    """

    def __init__(self, ishape, mat, adjoint=False):
        self.mat = mat
        self.adjoint = adjoint

        oshape = _get_matmul_oshape(ishape, mat.shape, adjoint)

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        xp = device.xp
        mat = backend.to_device(self.mat, device)
        with device:
            if self.adjoint:
                mat = xp.conj(mat).swapaxes(-1, -2)

            return xp.matmul(mat, input)

    def _adjoint_linop(self):
        sum_axes = _get_matmul_adjoint_sum_axes(
            self.oshape, self.ishape, self.mat.shape
        )

        M = MatMul(self.oshape, self.mat, adjoint=not self.adjoint)
        S = Sum(M.oshape, sum_axes)
        R = Reshape(self.ishape, S.oshape)
        return R * S * M


def _get_right_matmul_oshape(ishape, mshape, adjoint):
    ishape_exp, mshape_exp = util._expand_shapes(ishape, mshape)
    if adjoint:
        mshape_exp[-1], mshape_exp[-2] = mshape_exp[-2], mshape_exp[-1]

    max_ndim = max(len(ishape), len(mshape))
    oshape = []
    for i, m, d in zip(ishape_exp[:-2], mshape_exp[:-2], range(max_ndim - 2)):
        if not (i == m or i == 1 or m == 1):
            raise ValueError(
                "Invalid shapes: {ishape}, {mshape}.".format(
                    ishape=ishape, mshape=mshape
                )
            )

        oshape.append(max(i, m))

    if ishape_exp[-1] != mshape_exp[-2]:
        raise ValueError(
            "Invalid shapes: {ishape}, {mshape}.".format(
                ishape=ishape, mshape=mshape
            )
        )

    oshape += [ishape_exp[-2], mshape_exp[-1]]

    return oshape


class RightMatMul(Linop):
    """Matrix multiplication on the right.

    Args:
        ishape (tuple of ints): Input shape.
            It must be able to broadcast with mat.shape.
        mat (array): Matrix of shape [..., m, n]
        adjoint (bool): Toggle adjoint.
            If True, performs conj(mat).swapaxes(-1, -2)
            before performing matrix multiplication.

    """

    def __init__(self, ishape, mat, adjoint=False):
        self.mat = mat
        self.adjoint = adjoint

        oshape = _get_right_matmul_oshape(ishape, mat.shape, adjoint)

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        xp = device.xp
        mat = backend.to_device(self.mat, device)
        with device:
            if self.adjoint:
                mat = xp.conj(mat).swapaxes(-1, -2)

            return xp.matmul(input, mat)

    def _adjoint_linop(self):
        sum_axes = _get_matmul_adjoint_sum_axes(
            self.oshape, self.ishape, self.mat.shape
        )

        M = RightMatMul(self.oshape, self.mat, adjoint=not self.adjoint)
        S = Sum(M.oshape, sum_axes)
        R = Reshape(self.ishape, S.oshape)
        return R * S * M


def _get_multiply_oshape(ishape, mshape):
    ishape_exp, mshape_exp = util._expand_shapes(ishape, mshape)
    max_ndim = max(len(ishape), len(mshape))
    oshape = []
    for i, m, d in zip(ishape_exp, mshape_exp, range(max_ndim)):
        if not (i == m or i == 1 or m == 1):
            raise ValueError(
                "Invalid shapes: {ishape}, {mshape}.".format(
                    ishape=ishape, mshape=mshape
                )
            )

        oshape.append(max(i, m))

    return oshape


def _get_multiply_adjoint_sum_axes(oshape, ishape, mshape):
    ishape_exp, mshape_exp = util._expand_shapes(ishape, mshape)
    max_ndim = max(len(ishape), len(mshape))
    sum_axes = []
    for i, m, o, d in zip(ishape_exp, mshape_exp, oshape, range(max_ndim)):
        if i == 1 and (m != 1 or o != 1):
            sum_axes.append(d)

    return sum_axes


class Multiply(Linop):
    """Multiplication linear operator.

    Args:
        ishape (tuple of ints): Input shape.
        mult (array): Array to multiply.

    """

    def __init__(self, ishape, mult, conj=False):
        self.mult = mult
        self.conj = conj
        if np.isscalar(mult):
            self.mshape = [1]
        else:
            self.mshape = mult.shape

        oshape = _get_multiply_oshape(ishape, self.mshape)
        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        xp = device.xp
        with device:
            if np.isscalar(self.mult):
                if self.mult == 1:
                    return input

                mult = self.mult
                if self.conj:
                    mult = mult.conjugate()

            else:
                mult = backend.to_device(self.mult, backend.get_device(input))
                if self.conj:
                    mult = xp.conj(mult)

            return input * mult

    def _adjoint_linop(self):
        sum_axes = _get_multiply_adjoint_sum_axes(
            self.oshape, self.ishape, self.mshape
        )

        M = Multiply(self.oshape, self.mult, conj=not self.conj)
        S = Sum(M.oshape, sum_axes)
        R = Reshape(self.ishape, S.oshape)
        return R * S * M


class Interpolate(Linop):
    """Interpolation linear operator.

    Args:
        ishape (tuple of ints): Input shape = batch_shape + grd_shape
        coord (array): Coordinates, values from - nx / 2 to nx / 2 - 1.
                ndim can only be 1, 2 or 3, of shape pts_shape + [ndim]
        width (float): Width of interp. kernel in grid size.
        kernel (str): Interpolation kernel, {'spline', 'kaiser_bessel'}.
        param (float): Kernel parameter.

    See Also:
        :func:`sigpy.interpolate`

    """

    def __init__(self, ishape, coord, kernel="spline", width=2, param=1):
        ndim = coord.shape[-1]
        oshape = list(ishape[:-ndim]) + list(coord.shape[:-1])

        self.coord = coord
        self.width = width
        self.kernel = kernel
        self.param = param

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            coord = backend.to_device(self.coord, device)
            return interp.interpolate(
                input,
                coord,
                kernel=self.kernel,
                width=self.width,
                param=self.param,
            )

    def _adjoint_linop(self):
        return Gridding(
            self.ishape,
            self.coord,
            kernel=self.kernel,
            width=self.width,
            param=self.param,
        )


class Gridding(Linop):
    """Gridding linear operator.

    Args:
        oshape (tuple of ints): Output shape = batch_shape + pts_shape
        ishape (tuple of ints): Input shape = batch_shape + grd_shape
        coord (array): Coordinates, values from - nx / 2 to nx / 2 - 1.
                ndim can only be 1, 2 or 3. of shape pts_shape + [ndim]
        width (float): Width of interp. kernel in grid size.
        kernel (str): Interpolation kernel, {'spline', 'kaiser_bessel'}.
        param (float): Kernel parameter.

    See Also:
        :func:`sigpy.gridding`

    """

    def __init__(self, oshape, coord, kernel="spline", width=2, param=1):
        ndim = coord.shape[-1]
        ishape = list(oshape[:-ndim]) + list(coord.shape[:-1])

        self.coord = coord
        self.kernel = kernel
        self.width = width
        self.param = param

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            coord = backend.to_device(self.coord, device)
            return interp.gridding(
                input,
                coord,
                self.oshape,
                kernel=self.kernel,
                width=self.width,
                param=self.param,
            )

    def _adjoint_linop(self):
        return Interpolate(
            self.oshape,
            self.coord,
            kernel=self.kernel,
            width=self.width,
            param=self.param,
        )


class Resize(Linop):
    """Resize linear operator.

    Args:
        oshape (tuple of int): Output shape.
        ishape (tuple of int): Input shape
    """

    def __init__(self, oshape, ishape, ishift=None, oshift=None):
        self.ishift = ishift
        self.oshift = oshift

        super().__init__(oshape, ishape)

    def _apply(self, input):
        with backend.get_device(input):
            return util.resize(
                input, self.oshape, ishift=self.ishift, oshift=self.oshift
            )

    def _adjoint_linop(self):
        return Resize(
            self.ishape, self.oshape, ishift=self.oshift, oshift=self.ishift
        )


class Flip(Linop):
    """Flip linear operator.

    Args:
        shape (tuple of int): Input shape
    """

    def __init__(self, shape, axes=None):
        self.axes = axes

        super().__init__(shape, shape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            return util.flip(input, self.axes)

    def _adjoint_linop(self):
        return self


class Downsample(Linop):
    """Downsampling linear operator.

    Args:
        ishape (tuple of ints): Input shape.
        factor (tuple of ints): Downsampling factor.
        shift (None of tuple of ints): Shifts before down-sampling.
    """

    def __init__(self, ishape, factors, shift=None):
        self.factors = factors

        if shift is None:
            shift = [0] * len(ishape)

        self.shift = shift
        oshape = [
            ((i - s + f - 1) // f) for i, f, s in zip(ishape, factors, shift)
        ]

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            return util.downsample(input, self.factors, shift=self.shift)

    def _adjoint_linop(self):
        return Upsample(self.ishape, self.factors, shift=self.shift)


class Upsample(Linop):
    """Upsampling linear operator.

    Args:
        ishape (tuple of ints): Input shape.
        factor (tuple of ints): Upsampling factor.
        shift (None of tuple of ints): Shifts before up-sampling.

    """

    def __init__(self, oshape, factors, shift=None):
        self.factors = factors

        if shift is None:
            shift = [0] * len(oshape)

        self.shift = shift
        ishape = [
            ((i - s + f - 1) // f) for i, f, s in zip(oshape, factors, shift)
        ]

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            return util.upsample(
                input, self.oshape, self.factors, shift=self.shift
            )

    def _adjoint_linop(self):
        return Downsample(self.oshape, self.factors, shift=self.shift)


class Circshift(Linop):
    """Circular shift linear operator.

    Args:
        shape (tuple of ints): Input/output shape.
        shift (tuple of ints): Shifts.
        axes (None or tuple of ints): Axes to perform circular shift.

    """

    def __init__(self, shape, shift, axes=None):
        self.axes = axes
        self.shift = shift
        self.ishift = [-s for s in self.shift]

        super().__init__(shape, shape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            return util.circshift(input, self.shift, self.axes)

    def _adjoint_linop(self):
        return Circshift(self.ishape, [-s for s in self.shift], axes=self.axes)

    def _normal_linop(self):
        return Identity(self.ishape)


class Wavelet(Linop):
    """Wavelet transform linear operator.

    Currently only has CPU implementation. GPU inputs will be copied to CPU,
    and back to compute on CPU.

    Args:
        ishape (tuple of int): Input shape.
        axes (None or tuple of int): Axes to perform wavelet transform.
        wave_name (str): Wavelet name.
        level (None or int): Number of wavelet levels.
    """

    def __init__(self, ishape, axes=None, wave_name="db4", level=None):
        self.wave_name = wave_name
        self.axes = axes
        self.level = level
        oshape, _ = wavelet.get_wavelet_shape(
            ishape, wave_name=wave_name, axes=axes, level=level
        )

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            return wavelet.fwt(
                input,
                wave_name=self.wave_name,
                axes=self.axes,
                level=self.level,
            )

    def _adjoint_linop(self):
        return InverseWavelet(
            self.ishape,
            axes=self.axes,
            wave_name=self.wave_name,
            level=self.level,
        )


class InverseWavelet(Linop):
    """Inverse wavelet transform linear operator.

    Currently only has CPU implementation. GPU inputs will be copied to CPU,
    and back to compute on CPU.

    Args:
        oshape (tuple of int): Output shape.
        axes (None or tuple of int): Axes to perform wavelet transform.
        wave_name (str): Wavelet name.
        level (None or int): Number of wavelet levels.
    """

    def __init__(self, oshape, axes=None, wave_name="db4", level=None):
        self.wave_name = wave_name
        self.axes = axes
        self.level = level
        ishape, self.coeff_slices = wavelet.get_wavelet_shape(
            oshape, wave_name=wave_name, axes=axes, level=level
        )
        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            return wavelet.iwt(
                input,
                self.oshape,
                self.coeff_slices,
                wave_name=self.wave_name,
                axes=self.axes,
                level=self.level,
            )

    def _adjoint_linop(self):
        return Wavelet(
            self.oshape,
            axes=self.axes,
            wave_name=self.wave_name,
            level=self.level,
        )


class Sum(Linop):
    """Sum linear operator. Accumulate axes by summing.

    Args:
        ishape (tuple of ints): Input shape.
        axes (tuple of ints): Axes to sum over.
    """

    def __init__(self, ishape, axes):
        self.axes = tuple(i % len(ishape) for i in axes)
        oshape = [ishape[i] for i in range(len(ishape)) if i not in self.axes]

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        xp = device.xp
        with device:
            return xp.sum(input, axis=self.axes)

    def _adjoint_linop(self):
        return Tile(self.ishape, self.axes)


class Tile(Linop):
    """Tile linear operator.

    Args:
        oshape (tuple of ints): Output shape.
        axes (tuple of ints): Axes to tile.

    """

    def __init__(self, oshape, axes):
        self.axes = tuple(a % len(oshape) for a in axes)
        ishape = [oshape[d] for d in range(len(oshape)) if d not in self.axes]
        self.expanded_ishape = []
        self.reps = []
        for d in range(len(oshape)):
            if d in self.axes:
                self.expanded_ishape.append(1)
                self.reps.append(oshape[d])
            else:
                self.expanded_ishape.append(oshape[d])
                self.reps.append(1)

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        xp = device.xp
        with device:
            return xp.tile(input.reshape(self.expanded_ishape), self.reps)

    def _adjoint_linop(self):
        return Sum(self.oshape, self.axes)


class ArrayToBlocks(Linop):
    """Extract blocks from an array in a sliding window manner.

    Args:
        ishape (array): input array of shape [..., N_1, ..., N_D]
        blk_shape (tuple): block shape of length D, with D <= 4.
        blk_strides (tuple): block strides of length D.

    See Also:
        :func:`sigpy.block.array_to_blocks`

    """

    def __init__(self, ishape, blk_shape, blk_strides):
        self.blk_shape = blk_shape
        self.blk_strides = blk_strides
        D = len(blk_shape)
        num_blks = [
            (i - b + s) // s
            for i, b, s in zip(ishape[-D:], blk_shape, blk_strides)
        ]
        oshape = list(ishape[:-D]) + num_blks + list(blk_shape)

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            return block.array_to_blocks(
                input, self.blk_shape, self.blk_strides
            )

    def _adjoint_linop(self):
        return BlocksToArray(self.ishape, self.blk_shape, self.blk_strides)

    def _normal_linop(self):
        return Identity(self.ishape)


class BlocksToArray(Linop):
    """Accumulate blocks into an array in a sliding window manner.

    Args:
        oshape (tuple): output shape.
        blk_shape (tuple): block shape of length D.
        blk_strides (tuple): block strides of length D.

    Returns:
        array: array of shape oshape.

    """

    def __init__(self, oshape, blk_shape, blk_strides):
        self.blk_shape = blk_shape
        self.blk_strides = blk_strides
        D = len(blk_shape)
        num_blks = [
            (i - b + s) // s
            for i, b, s in zip(oshape[-D:], blk_shape, blk_strides)
        ]
        ishape = list(oshape[:-D]) + num_blks + list(blk_shape)

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            return block.blocks_to_array(
                input, self.oshape, self.blk_shape, self.blk_strides
            )

    def _adjoint_linop(self):
        return ArrayToBlocks(self.oshape, self.blk_shape, self.blk_strides)

    def _normal_linop(self):
        return Identity(self.ishape)


def Gradient(ishape, axes=None):
    import warnings

    warnings.warn(
        "Gradient Linop is renamed to FiniteDifference, "
        "Please switch to use FiniteDifference.",
        category=DeprecationWarning,
    )

    return FiniteDifference(ishape, axes=axes)


def FiniteDifference(ishape, axes=None):
    """Linear operator that computes finite difference gradient.

    Args:
        ishape (tuple of ints): Input shape.
        axes (tuple or list): Axes to circularly shift. All axes are used if
            None.

    """
    Id = Identity(ishape)
    ndim = len(ishape)
    axes = util._normalize_axes(axes, ndim)
    linops = []
    for i in axes:
        D = Id - Circshift(ishape, [1], axes=[i])
        R = Reshape([1] + list(ishape), ishape)
        linops.append(R * D)

    G = Vstack(linops, axis=0)

    return G


class NUFFT(Linop):
    """NUFFT linear operator.

    Args:
        ishape (tuple of int): Input shape.
        coord (array): Coordinates, with values [-ishape / 2, ishape / 2]
        oversamp (float): Oversampling factor.
        width (float): Kernel width.
        toeplitz (bool): Use toeplitz PSF to evaluate normal operator.

    """

    def __init__(self, ishape, coord, oversamp=1.25, width=4, toeplitz=False):
        self.coord = coord
        self.oversamp = oversamp
        self.width = width
        self.toeplitz = toeplitz

        ndim = coord.shape[-1]

        oshape = list(ishape[:-ndim]) + list(coord.shape[:-1])

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            coord = backend.to_device(self.coord, device)
            return fourier.nufft(
                input, coord, oversamp=self.oversamp, width=self.width
            )

    def _adjoint_linop(self):
        return NUFFTAdjoint(
            self.ishape, self.coord, oversamp=self.oversamp, width=self.width
        )

    def _normal_linop(self):
        if self.toeplitz is False:
            return self.H * self

        ndim = self.coord.shape[-1]
        psf = fourier.toeplitz_psf(
            self.coord, self.ishape, self.oversamp, self.width
        )

        fft_axes = tuple(range(-1, -(ndim + 1), -1))

        R = Resize(psf.shape, self.ishape)
        F = FFT(psf.shape, axes=fft_axes)
        P = Multiply(psf.shape, psf)
        T = R.H * F.H * P * F * R

        return T


class NUFFTAdjoint(Linop):
    """NUFFT adjoint linear operator.

    Args:
        oshape (tuple of int): Output shape
        coord (array): Coordinates, with values [-ishape / 2, ishape / 2]
        oversamp (float): Oversampling factor.
        width (float): Kernel width.

    """

    def __init__(self, oshape, coord, oversamp=1.25, width=4):
        self.coord = coord
        self.oversamp = oversamp
        self.width = width

        ndim = coord.shape[-1]

        ishape = list(oshape[:-ndim]) + list(coord.shape[:-1])

        super().__init__(oshape, ishape)

    def _apply(self, input):
        device = backend.get_device(input)
        with device:
            coord = backend.to_device(self.coord, device)
            return fourier.nufft_adjoint(
                input,
                coord,
                self.oshape,
                oversamp=self.oversamp,
                width=self.width,
            )

    def _adjoint_linop(self):
        return NUFFT(
            self.oshape, self.coord, oversamp=self.oversamp, width=self.width
        )


class ConvolveData(Linop):
    r"""Convolution operator for data arrays.

    Args:
        data_shape (tuple of ints): data array shape:
            :math:`[\ldots, m_1, \ldots, m_D]` if multi_channel is False,
            :math:`[\ldots, c_i, m_1, \ldots, m_D]` otherwise.
        filt (array): filter array of shape:
            :math:`[n_1, \ldots, n_D]` if multi_channel is False
            :math:`[c_o, c_i, n_1, \ldots, n_D]` otherwise.
        mode (str): {'full', 'valid'}.
        strides (None or tuple of ints): convolution strides of length D.
        multi_channel (bool): specify if input/output has multiple channels.

    """

    def __init__(
        self, data_shape, filt, mode="full", strides=None, multi_channel=False
    ):
        self.filt = filt
        self.mode = mode
        self.strides = strides
        self.multi_channel = multi_channel

        D, b, B, m, n, s, c_i, c_o, p = conv._get_convolve_params(
            data_shape, filt.shape, mode, strides, multi_channel
        )

        if multi_channel:
            output_shape = b + (c_o,) + p
        else:
            output_shape = b + p

        super().__init__(output_shape, data_shape)

    def _apply(self, input):
        device = backend.get_device(input)
        filt = backend.to_device(self.filt, device)
        with device:
            return conv.convolve(
                input,
                filt,
                mode=self.mode,
                strides=self.strides,
                multi_channel=self.multi_channel,
            )

    def _adjoint_linop(self):
        return ConvolveDataAdjoint(
            self.ishape,
            self.filt,
            mode=self.mode,
            strides=self.strides,
            multi_channel=self.multi_channel,
        )


class ConvolveDataAdjoint(Linop):
    r"""Adjoint convolution operator for data arrays.

    Args:
        data_shape (tuple of ints): data array shape:
            :math:`[\ldots, m_1, \ldots, m_D]` if multi_channel is False,
            :math:`[\ldots, c_i, m_1, \ldots, m_D]` otherwise.
        filt (array): filter array of shape:
            :math:`[n_1, \ldots, n_D]` if multi_channel is False
            :math:`[c_o, c_i, n_1, \ldots, n_D]` otherwise.
        mode (str): {'full', 'valid'}.
        strides (None or tuple of ints): convolution strides of length D.
        multi_channel (bool): specify if input/output has multiple channels.

    """

    def __init__(
        self, data_shape, filt, mode="full", strides=None, multi_channel=False
    ):
        self.filt = filt
        self.mode = mode
        self.strides = strides
        self.multi_channel = multi_channel

        D, b, B, m, n, s, c_i, c_o, p = conv._get_convolve_params(
            data_shape, filt.shape, mode, strides, multi_channel
        )

        if multi_channel:
            output_shape = b + (c_o,) + p
        else:
            output_shape = b + p

        super().__init__(data_shape, output_shape)

    def _apply(self, input):
        device = backend.get_device(input)
        filt = backend.to_device(self.filt, device)
        with device:
            return conv.convolve_data_adjoint(
                input,
                filt,
                self.oshape,
                mode=self.mode,
                strides=self.strides,
                multi_channel=self.multi_channel,
            )

    def _adjoint_linop(self):
        return ConvolveData(
            self.oshape,
            self.filt,
            mode=self.mode,
            strides=self.strides,
            multi_channel=self.multi_channel,
        )


class ConvolveFilter(Linop):
    r"""Convolution operator for filter arrays.

    Args:
        filt_shape (tuple of ints): filter array shape:
            :math:`[n_1, \ldots, n_D]` if multi_channel is False
            :math:`[c_o, c_i, n_1, \ldots, n_D]` otherwise.
        data (array): data array of shape:
            :math:`[\ldots, m_1, \ldots, m_D]` if multi_channel is False,
            :math:`[\ldots, c_i, m_1, \ldots, m_D]` otherwise.
        mode (str): {'full', 'valid'}.
        strides (None or tuple of ints): convolution strides of length D.
        multi_channel (bool): specify if input/output has multiple channels.

    """

    def __init__(
        self, filt_shape, data, mode="full", strides=None, multi_channel=False
    ):
        self.data = data
        self.mode = mode
        self.strides = strides
        self.multi_channel = multi_channel

        D, b, B, m, n, s, c_i, c_o, p = conv._get_convolve_params(
            data.shape, filt_shape, mode, strides, multi_channel
        )

        if multi_channel:
            output_shape = b + (c_o,) + p
        else:
            output_shape = b + p

        super().__init__(output_shape, filt_shape)

    def _apply(self, input):
        device = backend.get_device(input)
        data = backend.to_device(self.data, device)
        with device:
            return conv.convolve(
                data,
                input,
                mode=self.mode,
                strides=self.strides,
                multi_channel=self.multi_channel,
            )

    def _adjoint_linop(self):
        return ConvolveFilterAdjoint(
            self.ishape,
            self.data,
            mode=self.mode,
            strides=self.strides,
            multi_channel=self.multi_channel,
        )


class ConvolveFilterAdjoint(Linop):
    r"""Adjoint convolution operator for filter arrays.

    Args:
        filt_shape (tuple of ints): filter array shape:
            :math:`[n_1, \ldots, n_D]` if multi_channel is False
            :math:`[c_o, c_i, n_1, \ldots, n_D]` otherwise.
        data (array): data array of shape:
            :math:`[\ldots, m_1, \ldots, m_D]` if multi_channel is False,
            :math:`[\ldots, c_i, m_1, \ldots, m_D]` otherwise.
        mode (str): {'full', 'valid'}.
        strides (None or tuple of ints): convolution strides of length D.
        multi_channel (bool): specify if input/output has multiple channels.

    """

    def __init__(
        self, filt_shape, data, mode="full", strides=None, multi_channel=False
    ):
        self.data = data
        self.mode = mode
        self.strides = strides
        self.multi_channel = multi_channel

        D, b, B, m, n, s, c_i, c_o, p = conv._get_convolve_params(
            data.shape, filt_shape, mode, strides, multi_channel
        )

        if multi_channel:
            output_shape = b + (c_o,) + p
        else:
            output_shape = b + p

        super().__init__(filt_shape, output_shape)

    def _apply(self, input):
        device = backend.get_device(input)
        data = backend.to_device(self.data, device)
        with device:
            return conv.convolve_filter_adjoint(
                input,
                data,
                self.oshape,
                mode=self.mode,
                strides=self.strides,
                multi_channel=self.multi_channel,
            )

    def _adjoint_linop(self):
        return ConvolveFilter(
            self.oshape,
            self.data,
            mode=self.mode,
            strides=self.strides,
            multi_channel=self.multi_channel,
        )


class Slice(Linop):
    """Slice input with given index.

    Given input `input` and index `idx`, returns `input[idx]`.

    Args:
        ishape (tuple of ints): Input shape.
        idx (slice or tuple of slices): Index.

    """

    def __init__(self, ishape, idx):
        self.idx = idx
        oshape = np.empty(ishape)[idx].shape
        super().__init__(oshape, ishape)

    def _apply(self, input):
        return input[self.idx]

    def _adjoint_linop(self):
        return Embed(self.ishape, self.idx)


class Embed(Linop):
    """Embed input into a zero array with the given shape and index.

    Given input `input` and index `idx`,
    returns output with `output[idx] = input`.

    Args:
        oshape (tuple of ints): output shape.
        idx (slice or tuple of slices): Index.

    """

    def __init__(self, oshape, idx):
        self.idx = idx
        ishape = np.empty(oshape)[idx].shape
        super().__init__(oshape, ishape)

    def _apply(self, input):
        output = np.zeros(self.oshape, dtype=input.dtype)
        output[self.idx] = input
        return output

    def _adjoint_linop(self):
        return Slice(self.oshape, self.idx)



# -*- coding: utf-8 -*-
"""Block reshape functions.

"""
import numpy as np

from sigpy import backend, config, util

__all__ = ["array_to_blocks", "blocks_to_array"]


def array_to_blocks(input, blk_shape, blk_strides):
    """Extract blocks from an array in a sliding window manner.

    Args:
        input (array): input array of shape [..., N_1, ..., N_ndim]
        blk_shape (tuple): block shape of length ndim, with ndim={1, 2, 3}.
        blk_strides (tuple): block strides of length ndim.

    Returns:
        array: array of shape [...] + num_blks + blk_shape, where
            num_blks = (N - blk_shape + blk_strides) // blk_strides.

    Example:

        >>> input = np.array([0, 1, 2, 3, 4, 5])
        >>> print(array_to_blocks(input, [2], [2]))
        [[0, 1],
         [2, 3],
         [4, 5]]

    """
    if len(blk_shape) != len(blk_strides):
        raise ValueError("blk_shape must have the same length as blk_strides.")

    ndim = len(blk_shape)
    num_blks = [
        (i - b + s) // s
        for i, b, s in zip(input.shape[-ndim:], blk_shape, blk_strides)
    ]
    batch_shape = list(input.shape[:-ndim])
    batch_size = util.prod(batch_shape)
    xp = backend.get_array_module(input)
    output = xp.zeros(
        [batch_size] + num_blks + list(blk_shape), dtype=input.dtype
    )
    input = input.reshape([batch_size] + list(input.shape[-ndim:]))

    if ndim == 1:
        if xp == np:
            _array_to_blocks1(
                output,
                input,
                batch_size,
                blk_shape[-1],
                blk_strides[-1],
                num_blks[-1],
            )
        else:  # pragma: no cover
            _array_to_blocks1_cuda(
                input,
                batch_size,
                blk_shape[-1],
                blk_strides[-1],
                num_blks[-1],
                output,
                size=output.size,
            )
    elif ndim == 2:
        if xp == np:
            _array_to_blocks2(
                output,
                input,
                batch_size,
                blk_shape[-1],
                blk_shape[-2],
                blk_strides[-1],
                blk_strides[-2],
                num_blks[-1],
                num_blks[-2],
            )
        else:  # pragma: no cover
            _array_to_blocks2_cuda(
                input,
                batch_size,
                blk_shape[-1],
                blk_shape[-2],
                blk_strides[-1],
                blk_strides[-2],
                num_blks[-1],
                num_blks[-2],
                output,
                size=output.size,
            )
    elif ndim == 3:
        if xp == np:
            _array_to_blocks3(
                output,
                input,
                batch_size,
                blk_shape[-1],
                blk_shape[-2],
                blk_shape[-3],
                blk_strides[-1],
                blk_strides[-2],
                blk_strides[-3],
                num_blks[-1],
                num_blks[-2],
                num_blks[-3],
            )
        else:  # pragma: no cover
            _array_to_blocks3_cuda(
                input,
                batch_size,
                blk_shape[-1],
                blk_shape[-2],
                blk_shape[-3],
                blk_strides[-1],
                blk_strides[-2],
                blk_strides[-3],
                num_blks[-1],
                num_blks[-2],
                num_blks[-3],
                output,
                size=output.size,
            )
    else:
        raise ValueError("Only support ndim=1, 2, or 3, got {}".format(ndim))

    return output.reshape(batch_shape + num_blks + list(blk_shape))


def blocks_to_array(input, oshape, blk_shape, blk_strides):
    """Accumulate blocks into an array in a sliding window manner.

    Args:
        input (array): input array of shape [...] + num_blks + blk_shape
        oshape (tuple): output shape.
        blk_shape (tuple): block shape of length ndim.
        blk_strides (tuple): block strides of length ndim.

    Returns:
        array: array of shape oshape.

    """
    if len(blk_shape) != len(blk_strides):
        raise ValueError("blk_shape must have the same length as blk_strides.")

    ndim = len(blk_shape)
    num_blks = input.shape[-(2 * ndim) : -ndim]
    batch_shape = list(oshape[:-ndim])
    batch_size = util.prod(batch_shape)
    xp = backend.get_array_module(input)
    output = xp.zeros([batch_size] + list(oshape[-ndim:]), dtype=input.dtype)
    input = input.reshape([batch_size] + list(input.shape[-2 * ndim :]))

    if ndim == 1:
        if xp == np:
            _blocks_to_array1(
                output,
                input,
                batch_size,
                blk_shape[-1],
                blk_strides[-1],
                num_blks[-1],
            )
        else:  # pragma: no cover
            _blocks_to_array1_cuda(
                input,
                batch_size,
                blk_shape[-1],
                blk_strides[-1],
                num_blks[-1],
                output,
                size=output.size,
            )
    elif ndim == 2:
        if xp == np:
            _blocks_to_array2(
                output,
                input,
                batch_size,
                blk_shape[-1],
                blk_shape[-2],
                blk_strides[-1],
                blk_strides[-2],
                num_blks[-1],
                num_blks[-2],
            )
        else:  # pragma: no cover
            _blocks_to_array2_cuda(
                input,
                batch_size,
                blk_shape[-1],
                blk_shape[-2],
                blk_strides[-1],
                blk_strides[-2],
                num_blks[-1],
                num_blks[-2],
                output,
                size=output.size,
            )
    elif ndim == 3:
        if xp == np:
            _blocks_to_array3(
                output,
                input,
                batch_size,
                blk_shape[-1],
                blk_shape[-2],
                blk_shape[-3],
                blk_strides[-1],
                blk_strides[-2],
                blk_strides[-3],
                num_blks[-1],
                num_blks[-2],
                num_blks[-3],
            )
        else:  # pragma: no cover
            _blocks_to_array3_cuda(
                input,
                batch_size,
                blk_shape[-1],
                blk_shape[-2],
                blk_shape[-3],
                blk_strides[-1],
                blk_strides[-2],
                blk_strides[-3],
                num_blks[-1],
                num_blks[-2],
                num_blks[-3],
                output,
                size=output.size,
            )
    else:
        raise ValueError("Only support ndim=1, 2, or 3, got {}".format(ndim))

    return output.reshape(oshape)


@util.fallback_jit(nopython=True, cache=True)  # pragma: no cover
def _array_to_blocks1(output, input, batch_size, Bx, Sx, Nx):
    for b in range(batch_size):
        for nx in range(Nx):
            for bx in range(Bx):
                ix = nx * Sx + bx
                if ix < input.shape[-1]:
                    output[b, nx, bx] = input[b, ix]


@util.fallback_jit(nopython=True, cache=True)  # pragma: no cover
def _array_to_blocks2(output, input, batch_size, Bx, By, Sx, Sy, Nx, Ny):
    for b in range(batch_size):
        for ny in range(Ny):
            for nx in range(Nx):
                for by in range(By):
                    for bx in range(Bx):
                        iy = ny * Sy + by
                        ix = nx * Sx + bx
                        if ix < input.shape[-1] and iy < input.shape[-2]:
                            output[b, ny, nx, by, bx] = input[b, iy, ix]


@util.fallback_jit(nopython=True, cache=True)  # pragma: no cover
def _array_to_blocks3(
    output, input, batch_size, Bx, By, Bz, Sx, Sy, Sz, Nx, Ny, Nz
):
    for b in range(batch_size):
        for nz in range(Nz):
            for ny in range(Ny):
                for nx in range(Nx):
                    for bz in range(Bz):
                        for by in range(By):
                            for bx in range(Bx):
                                iz = nz * Sz + bz
                                iy = ny * Sy + by
                                ix = nx * Sx + bx
                                if (
                                    ix < input.shape[-1]
                                    and iy < input.shape[-2]
                                    and iz < input.shape[-3]
                                ):
                                    output[b, nz, ny, nx, bz, by, bx] = input[
                                        b, iz, iy, ix
                                    ]


@util.fallback_jit(nopython=True, cache=True)  # pragma: no cover
def _blocks_to_array1(output, input, batch_size, Bx, Sx, Nx):
    for b in range(batch_size):
        for ix in range(output.shape[-1]):
            rx = ix % Sx
            for bx in range(rx, Bx, Sx):
                nx = (ix - bx) // Sx
                if nx >= 0 and nx < Nx:
                    output[b, ix] += input[b, nx, bx]


@util.fallback_jit(nopython=True, cache=True)  # pragma: no cover
def _blocks_to_array2(output, input, batch_size, Bx, By, Sx, Sy, Nx, Ny):
    for b in range(batch_size):
        for iy in range(output.shape[-2]):
            ry = iy % Sy
            for ix in range(output.shape[-1]):
                rx = ix % Sx
                for by in range(ry, By, Sy):
                    ny = (iy - by) // Sy
                    if ny >= 0 and ny < Ny:
                        for bx in range(rx, Bx, Sx):
                            nx = (ix - bx) // Sx
                            if nx >= 0 and nx < Nx:
                                output[b, iy, ix] += input[b, ny, nx, by, bx]


@util.fallback_jit(nopython=True, cache=True)  # pragma: no cover
def _blocks_to_array3(
    output, input, batch_size, Bx, By, Bz, Sx, Sy, Sz, Nx, Ny, Nz
):
    for b in range(batch_size):
        for iz in range(output.shape[-3]):
            for iy in range(output.shape[-2]):
                for ix in range(output.shape[-1]):
                    rz = iz % Sz
                    ry = iy % Sy
                    rx = ix % Sx
                    for bz in range(rz, Bz, Sz):
                        for by in range(ry, By, Sy):
                            for bx in range(rx, Bx, Sx):
                                nz = (iz - bz) // Sz
                                ny = (iy - by) // Sy
                                nx = (ix - bx) // Sx
                                if (
                                    nx >= 0
                                    and nx < Nx
                                    and ny >= 0
                                    and ny < Ny
                                    and nz >= 0
                                    and nz < Nz
                                ):
                                    output[b, iz, iy, ix] += input[
                                        b, nz, ny, nx, bz, by, bx
                                    ]


if config.cupy_enabled:  # pragma: no cover
    import cupy as cp

    _array_to_blocks1_cuda = cp.ElementwiseKernel(
        "raw T input, int32 batch_size, int32 Bx, int32 Sx, int32 Nx",
        "raw T output",
        """
        const int ndim = input.ndim;

        int b = i / Bx / Nx;
        i -= b * Bx * Nx;
        int nx = i / Bx;
        i -= nx * Bx;
        int bx = i;

        int ix = nx * Sx + bx;
        if (ix < input.shape()[ndim - 1]) {
            int input_idx[] = {b, ix};
            int output_idx[] = {b, nx, bx};
            output[output_idx] = input[input_idx];
        }
        """,
        name="_array_to_blocks1_cuda",
    )

    _array_to_blocks2_cuda = cp.ElementwiseKernel(
        "raw T input, int32 batch_size, int32 Bx, int32 By, "
        "int32 Sx, int32 Sy, int32 Nx, int32 Ny",
        "raw T output",
        """
        const int ndim = input.ndim;

        int b = i / Bx / By / Nx / Ny;
        i -= b * Bx * By * Nx * Ny;
        int ny = i / Bx / By / Nx;
        i -= ny * Bx * By * Nx;
        int nx = i / Bx / By;
        i -= nx * Bx * By;
        int by = i / Bx;
        i -= by * Bx;
        int bx = i;

        int iy = ny * Sy + by;
        int ix = nx * Sx + bx;
        if (ix < input.shape()[ndim - 1] && iy < input.shape()[ndim - 2]) {
            int input_idx[] = {b, iy, ix};
            int output_idx[] = {b, ny, nx, by, bx};
            output[output_idx] = input[input_idx];
        }
        """,
        name="_array_to_blocks2_cuda",
    )

    _array_to_blocks3_cuda = cp.ElementwiseKernel(
        "raw T input, int32 batch_size, int32 Bx, int32 By, int32 Bz, "
        "int32 Sx, int32 Sy, int32 Sz, int32 Nx, int32 Ny, int32 Nz",
        "raw T output",
        """
        const int ndim = input.ndim;

        int b = i / Bx / By / Bz / Nx / Ny / Nz;
        i -= b * Bx * By * Bz * Nx * Ny * Nz;
        int nz = i / Bx / By / Bz / Nx / Ny;
        i -= nz * Bx * By * Bz * Nx * Ny;
        int ny = i / Bx / By / Bz / Nx;
        i -= ny * Bx * By * Bz * Nx;
        int nx = i / Bx / By / Bz;
        i -= nx * Bx * By * Bz;
        int bz = i / Bx / By;
        i -= bz * Bx * By;
        int by = i / Bx;
        i -= by * Bx;
        int bx = i;

        int iz = nz * Sz + bz;
        int iy = ny * Sy + by;
        int ix = nx * Sx + bx;
        if (ix < input.shape()[ndim - 1] && iy < input.shape()[ndim - 2]
            && iz < input.shape()[ndim - 3]) {
            int input_idx[] = {b, iz, iy, ix};
            int output_idx[] = {b, nz, ny, nx, bz, by, bx};
            output[output_idx] = input[input_idx];
        }
        """,
        name="_array_to_blocks3_cuda",
    )

    _blocks_to_array1_cuda = cp.ElementwiseKernel(
        "raw T input, int32 batch_size, int32 Bx, int32 Sx, int32 Nx",
        "raw T output",
        """
        const int ndim = output.ndim;

        int Ix = output.shape()[ndim - 1];
        int b = i / Ix;
        i -= b * Ix;
        int ix = i;

        int rx = ix % Sx;
        for (int bx = rx; bx < Bx; bx += Sx){
            int nx = (ix - bx) / Sx;
            if ((nx >= 0) && (nx < Nx)) {
                int input_idx[] = {b, nx, bx};
                int output_idx[] = {b, ix};
                output[output_idx] += input[input_idx];
            }
        }
        """,
        name="_blocks_to_array1_cuda",
    )

    _blocks_to_array2_cuda = cp.ElementwiseKernel(
        "raw T input, int32 batch_size, "
        "int32 Bx, int32 By, int32 Sx, int32 Sy, "
        "int32 Nx, int32 Ny",
        "raw T output",
        """
        const int ndim = output.ndim;

        int Ix = output.shape()[ndim - 1];
        int Iy = output.shape()[ndim - 2];
        int b = i / Ix / Iy;
        i -= b * Ix * Iy;
        int iy = i / Ix;
        i -= iy * Ix;
        int ix = i;

        int ry = iy % Sy;
        int rx = ix % Sx;
        for (int by = ry; by < By; by += Sy){
            int ny = (iy - by) / Sy;
            if ((ny >= 0) && (ny < Ny)) {
                for (int bx = rx; bx < Bx; bx += Sx){
                    int nx = (ix - bx) / Sx;
                    if ((nx >= 0) && (nx < Nx)) {
                        int input_idx[] = {b, ny, nx, by, bx};
                        int output_idx[] = {b, iy, ix};
                        output[output_idx] += input[input_idx];
                    }
                }
            }
        }
        """,
        name="_blocks_to_array2_cuda",
    )

    _blocks_to_array3_cuda = cp.ElementwiseKernel(
        "raw T input, int32 batch_size, int32 Bx, int32 By, int32 Bz, "
        "int32 Sx, int32 Sy, int32 Sz, int32 Nx, int32 Ny, int32 Nz",
        "raw T output",
        """
        const int ndim = output.ndim;

        int Ix = output.shape()[ndim - 1];
        int Iy = output.shape()[ndim - 2];
        int Iz = output.shape()[ndim - 3];
        int b = i / Ix / Iy / Iz;
        i -= b * Ix * Iy * Iz;
        int iz = i / Ix / Iy;
        i -= iz * Ix * Iy;
        int iy = i / Ix;
        i -= iy * Ix;
        int ix = i;

        int rz = iz % Sz;
        int ry = iy % Sy;
        int rx = ix % Sx;
        for (int bz = rz; bz < Bz; bz += Sz){
            int nz = (iz - bz) / Sz;
            if ((nz >= 0) && (nz < Nz)) {
                for (int by = ry; by < By; by += Sy){
                    int ny = (iy - by) / Sy;
                    if ((ny >= 0) && (ny < Ny)) {
                        for (int bx = rx; bx < Bx; bx += Sx){
                            int nx = (ix - bx) / Sx;
                            if ((nx >= 0) && (nx < Nx)) {
                                int input_idx[] = {b, nz, ny, nx, bz, by, bx};
                                int output_idx[] = {b, iz, iy, ix};
                                output[output_idx] += input[input_idx];
                            }
                        }
                    }
                }
            }
        }
        """,
        name="_blocks_to_array3_cuda",
    )



"""Density compensation functions.

"""
from tqdm.auto import tqdm

import sigpy as sp

__all__ = ["pipe_menon_dcf"]


def pipe_menon_dcf(
    coord,
    img_shape=None,
    device=sp.cpu_device,
    max_iter=30,
    n=128,
    beta=8,
    width=4,
    show_pbar=True,
):
    r"""Compute Pipe Menon density compensation factor.

    Perform the following iteration:

    .. math::

        w = \frac{w}{|G^H G w|}

    with :math:`G` as the gridding operator.

    Args:
        coord (array): k-space coordinates.
        img_shape (None or list): Image shape.
        device (Device): computing device.
        max_iter (int): number of iterations.
        n (int): Kaiser-Bessel sampling numbers for gridding operator.
        beta (float): Kaiser-Bessel kernel parameter.
        width (float): Kaiser-Bessel kernel width.
        show_pbar (bool): show progress bar.

    Returns:
        array: density compensation factor.

    References:
        Pipe, James G., and Padmanabhan Menon.
        Sampling Density Compensation in MRI:
        Rationale and an Iterative Numerical Solution.
        Magnetic Resonance in Medicine 41, no. 1 (1999): 179–86.


    """
    device = sp.Device(device)
    xp = device.xp

    with device:
        w = xp.ones(coord.shape[:-1], dtype=coord.dtype)
        if img_shape is None:
            img_shape = sp.estimate_shape(coord)

        G = sp.linop.Gridding(
            img_shape, coord, param=beta, width=width, kernel="kaiser_bessel"
        )
        with tqdm(
            total=max_iter, desc="PipeMenonDCF", disable=not show_pbar
        ) as pbar:
            for it in range(max_iter):
                GHGw = G.H * G * w
                w /= xp.abs(GHGw)
                resid = xp.abs(GHGw - 1).max().item()

                pbar.set_postfix(resid="{0:.2E}".format(resid))
                pbar.update()

    return w



# -*- coding: utf-8 -*-
"""MRI RF excitation pulse design functions,
    including SLR and small tip spatial design
"""

import numpy as np
import scipy.linalg as linalg
import scipy.signal as signal

import sigpy as sp
from sigpy.mri.rf.util import dinf

__all__ = [
    "dzrf",
    "dzls",
    "msinc",
    "dzmp",
    "fmp",
    "dzlp",
    "b2rf",
    "b2a",
    "mag2mp",
    "ab2rf",
    "dz_gslider_b",
    "dz_gslider_rf",
    "root_flip",
    "dz_recursive_rf",
    "dz_hadamard_b",
    "calc_ripples",
]

""" Functions for SLR pulse design
    SLR algorithm simplifies the solution of the Bloch equations
    to the design of 2 polynomials
"""


def dzrf(
    n=64,
    tb=4,
    ptype="st",
    ftype="ls",
    d1=0.01,
    d2=0.01,
    cancel_alpha_phs=False,
):
    r"""Primary function for design of pulses using the SLR algorithm.

    Args:
        n (int): number of time points.
        tb (int): pulse time bandwidth product.
        ptype (string): pulse type, 'st' (small-tip excitation), 'ex' (pi/2
            excitation pulse), 'se' (spin-echo pulse), 'inv' (inversion), or
            'sat' (pi/2 saturation pulse).
        ftype (string): type of filter to use: 'ms' (sinc), 'pm'
            (Parks-McClellan equal-ripple), 'min' (minphase using factored pm),
            'max' (maxphase using factored pm), 'ls' (least squares).
        d1 (float): passband ripple level in :math:'M_0^{-1}'.
        d2 (float): stopband ripple level in :math:'M_0^{-1}'.
        cancel_alpha_phs (bool): For 'ex' pulses, absorb the alpha phase
            profile from beta's profile, so they cancel for a flatter
            total phase

    Returns:
        rf (array): designed RF pulse.

    References:
        Pauly, J., Le Roux, Patrick., Nishimura, D., and Macovski, A.(1991).
        Parameter Relations for the Shinnar-LeRoux Selective Excitation
        Pulse Design Algorithm.
        IEEE Transactions on Medical Imaging, Vol 10, No 1, 53-65.
    """

    [bsf, d1, d2] = calc_ripples(ptype, d1, d2)

    if ftype == "ms":  # sinc
        b = msinc(n, tb / 4)
    elif ftype == "pm":  # linphase
        b = dzlp(n, tb, d1, d2)
    elif ftype == "min":  # minphase
        b = dzmp(n, tb, d1, d2)
        b = b[::-1]
    elif ftype == "max":  # maxphase
        b = dzmp(n, tb, d1, d2)
    elif ftype == "ls":  # least squares
        b = dzls(n, tb, d1, d2)
    else:
        raise Exception('Filter type ("{}") is not recognized.'.format(ftype))

    if ptype == "st":
        rf = b
    elif ptype == "ex":
        b = bsf * b
        rf = b2rf(b, cancel_alpha_phs)
    else:
        b = bsf * b
        rf = b2rf(b)

    return rf


def calc_ripples(ptype="st", d1=0.01, d2=0.01):
    if ptype == "st":
        bsf = 1
    elif ptype == "ex":
        bsf = np.sqrt(1 / 2)
        d1 = np.sqrt(d1 / 2)
        d2 = d2 / np.sqrt(2)
    elif ptype == "se":
        bsf = 1
        d1 = d1 / 4
        d2 = np.sqrt(d2)
    elif ptype == "inv":
        bsf = 1
        d1 = d1 / 8
        d2 = np.sqrt(d2 / 2)
    elif ptype == "sat":
        bsf = np.sqrt(1 / 2)
        d1 = d1 / 2
        d2 = np.sqrt(d2)
    else:
        raise Exception('Pulse type ("{}") is not recognized.'.format(ptype))

    return bsf, d1, d2


# following functions are used to support dzrf


def dzls(n=64, tb=4, d1=0.01, d2=0.01):
    di = dinf(d1, d2)
    w = di / tb
    f = np.asarray([0, (1 - w) * (tb / 2), (1 + w) * (tb / 2), (n / 2)])
    f = f / (n / 2)
    m = [1, 1, 0, 0]
    w = [1, d1 / d2]

    h = signal.firls(n + 1, f, m, w)
    # shift the filter half a sample to make it symmetric, like in MATLAB
    c = np.exp(
        1j
        * 2
        * np.pi
        / (2 * (n + 1))
        * np.concatenate([np.arange(0, n / 2 + 1, 1), np.arange(-n / 2, 0, 1)])
    )
    h = np.real(sp.ifft(np.multiply(sp.fft(h, center=False), c), center=False))

    # lop off extra sample
    h = h[:n]

    return h


def dzmp(n=64, tb=4, d1=0.01, d2=0.01):
    n2 = 2 * n - 1
    di = 0.5 * dinf(2 * d1, 0.5 * d2 * d2)
    w = di / tb
    f = np.asarray([0, (1 - w) * (tb / 2), (1 + w) * (tb / 2), (n / 2)]) / n
    m = [1, 0]
    w = [1, 2 * d1 / (0.5 * d2 * d2)]

    hl = signal.remez(n2, f, m, w)

    h = fmp(hl)

    return h


def fmp(h):
    ll = np.size(h)
    lp = 128 * np.exp(np.ceil(np.log(ll) / np.log(2)) * np.log(2))
    padwidths = np.array([np.ceil((lp - ll) / 2), np.floor((lp - ll) / 2)])
    hp = np.pad(h, padwidths.astype(int), "constant")
    hpf = sp.fft(hp, norm=None)
    hpfs = hpf - np.min(np.real(hpf)) * 1.000001
    hpfmp = mag2mp(np.sqrt(np.abs(hpfs)))
    hpmp = sp.ifft(np.fft.ifftshift(np.conj(hpfmp)), center=False, norm=None)
    hmp = hpmp[: int((ll + 1) / 2)]

    return hmp


def dzlp(n=64, tb=4, d1=0.01, d2=0.01):
    di = dinf(d1, d2)
    w = di / tb
    f = np.asarray([0, (1 - w) * (tb / 2), (1 + w) * (tb / 2), (n / 2)]) / n
    m = [1, 0]
    w = [1, d1 / d2]

    h = signal.remez(n, f, m, w)

    return h


def msinc(n=64, m=1):
    x = np.arange(-n / 2, n / 2, 1) / (n / 2)
    snc = np.divide(
        np.sin(m * 2 * np.pi * x + 0.00001), (m * 2 * np.pi * x + 0.00001)
    )
    ms = np.multiply(snc, 0.54 + 0.46 * np.cos(np.pi * x))
    ms = ms * 4 * m / n

    return ms


def dz_gslider_b(
    n=128, g=5, gind=1, tb=4, d1=0.01, d2=0.01, phi=np.pi, shift=32
):
    r"""Design a g-slider pulse b

    Args:
        n (int): number of time points.
        g (int): number of sub-slices.
        gind (int): subslice index.
        tb (int): time bandwidth product.
        d1 (float): passband ripple level in :math:'M_0^{-1}'.
        d2 (float): stopband ripple level in :math:'M_0^{-1}'.
        phi (float): subslice phase.
        shift (int): n time points shift of pulse.

    Returns:
        b (array): SLR beta parameter.

    References:
        Setsompop, K. et al. 'High-resolution in vivo diffusion imaging of the
        human brain with generalized slice dithered enhanced resolution:
        Simultaneous multislice (gSlider-SMS). Magn. Reson. Med.79, 141–151
        (2018).
    """
    ftw = dinf(d1, d2) / tb  # fractional transition width of the slab profile

    if np.fmod(g, 2) and gind == int(np.ceil(g / 2)):  # centered sub-slice
        if g == 1:  # no sub-slices, as a sanity check
            b = dzls(n, tb, d1, d2)
        else:
            # Design 2 filters, to allow arbitrary phases on the subslice the
            # first is a wider notch filter with '0's where it the subslice
            # appears, and the second is the subslice. Multiply the subslice by
            # its phase and add the filters.
            f = np.asarray(
                [
                    0,
                    (1 / g - ftw) * (tb / 2),
                    (1 / g + ftw) * (tb / 2),
                    (1 - ftw) * (tb / 2),
                    (1 + ftw) * (tb / 2),
                    (n / 2),
                ]
            )
            f = f / (n / 2)
            m_notch = [0, 0, 1, 1, 0, 0]
            m_sub = [1, 1, 0, 0, 0, 0]
            w = [1, 1, d1 / d2]

            b_notch = signal.firls(n + 1, f, m_notch, w)  # the notched filter
            b_sub = signal.firls(n + 1, f, m_sub, w)  # the subslice filter
            # add them with the subslice phase
            b = np.add(b_notch, np.multiply(np.exp(1j * phi), b_sub))
            # shift the filter half a sample to make it symmetric,
            # like in MATLAB
            c = np.exp(
                1j
                * 2
                * np.pi
                / (2 * (n + 1))
                * np.concatenate(
                    [np.arange(0, n / 2 + 1, 1), np.arange(-n / 2, 0, 1)]
                )
            )
            b = sp.ifft(np.multiply(sp.fft(b, center=False), c), center=False)
            # lop off extra sample
            b = b[:n]

    else:
        # design filters for the slab and the subslice, hilbert xform them
        # to suppress their left bands,
        # then demodulate the result back to DC
        gcent = shift + (gind - g / 2 - 1 / 2) * tb / g
        if gind > 1 and gind < g:
            # separate transition bands for slab+slice
            f = np.asarray(
                [
                    0,
                    shift - (1 + ftw) * (tb / 2),
                    shift - (1 - ftw) * (tb / 2),
                    gcent - (tb / g / 2 + ftw * (tb / 2)),
                    gcent - (tb / g / 2 - ftw * (tb / 2)),
                    gcent + (tb / g / 2 - ftw * (tb / 2)),
                    gcent + (tb / g / 2 + ftw * (tb / 2)),
                    shift + (1 - ftw) * (tb / 2),
                    shift + (1 + ftw) * (tb / 2),
                    (n / 2),
                ]
            )
            f = f / (n / 2)
            m_notch = [0, 0, 1, 1, 0, 0, 1, 1, 0, 0]
            m_sub = [0, 0, 0, 0, 1, 1, 0, 0, 0, 0]
            w = [d1 / d2, 1, 1, 1, d1 / d2]
        elif gind == 1:
            # the slab and slice share a left transition band
            f = np.asarray(
                [
                    0,
                    shift - (1 + ftw) * (tb / 2),
                    shift - (1 - ftw) * (tb / 2),
                    gcent + (tb / g / 2 - ftw * (tb / 2)),
                    gcent + (tb / g / 2 + ftw * (tb / 2)),
                    shift + (1 - ftw) * (tb / 2),
                    shift + (1 + ftw) * (tb / 2),
                    (n / 2),
                ]
            )
            f = f / (n / 2)
            m_notch = [0, 0, 0, 0, 1, 1, 0, 0]
            m_sub = [0, 0, 1, 1, 0, 0, 0, 0]
            w = [d1 / d2, 1, 1, d1 / d2]
        elif gind == g:
            # the slab and slice share a right transition band
            f = np.asarray(
                [
                    0,
                    shift - (1 + ftw) * (tb / 2),
                    shift - (1 - ftw) * (tb / 2),
                    gcent - (tb / g / 2 + ftw * (tb / 2)),
                    gcent - (tb / g / 2 - ftw * (tb / 2)),
                    shift + (1 - ftw) * (tb / 2),
                    shift + (1 + ftw) * (tb / 2),
                    (n / 2),
                ]
            )
            f = f / (n / 2)
            m_notch = [0, 0, 1, 1, 0, 0, 0, 0]
            m_sub = [0, 0, 0, 0, 1, 1, 0, 0]
            w = [d1 / d2, 1, 1, d1 / d2]

        c = np.exp(
            1j
            * 2
            * np.pi
            / (2 * (n + 1))
            * np.concatenate(
                [np.arange(0, n / 2 + 1, 1), np.arange(-n / 2, 0, 1)]
            )
        )

        b_notch = signal.firls(n + 1, f, m_notch, w)  # the notched filter
        b_notch = sp.ifft(
            np.multiply(sp.fft(b_notch, center=False), c), center=False
        )
        b_notch = np.real(b_notch[:n])
        # hilbert transform to suppress negative passband
        b_notch = signal.hilbert(b_notch)

        b_sub = signal.firls(n + 1, f, m_sub, w)  # the sub-band filter
        b_sub = sp.ifft(
            np.multiply(sp.fft(b_sub, center=False), c), center=False
        )

        b_sub = np.real(b_sub[:n])
        # hilbert transform to suppress negative passband
        b_sub = signal.hilbert(b_sub)

        # add them with the subslice phase
        b = b_notch + np.exp(1j * phi) * b_sub

        # demodulate to DC
        c_shift = np.exp(-1j * 2 * np.pi / n * shift * np.arange(0, n, 1)) / 2
        c_shift *= np.exp(-1j * np.pi / n * shift)

        b = np.multiply(b, c_shift)

    return b


def dz_hadamard_b(n=128, g=5, gind=1, tb=4, d1=0.01, d2=0.01, shift=32):
    r"""Design a pulse with hadamard encoding

    Args:
        n (int): number of time points.
        g (int): order of the Hadamard matrix.
        gind (int): index of vector to use from Hadamard matrix for encoding.
        tb (int): time bandwidth product.
        d1 (float): passband ripple level in :math:'M_0^{-1}'.
        d2 (float): stopband ripple level in :math:'M_0^{-1}'.
        shift (int): n time points shift of pulse.

    Returns:
        b (array): SLR beta parameter.

    References:
            Souza, S.P., Szumowski, J., Dumoulin, C.L., Plewes, D.P. &
            Glover, G. 'Sima: Simultaneous multislice acquisition of MR images
            by hadamard - encoded excitation. J.Comput.Assist.Tomogr. 12,
            1026–1030(1988).
    """

    H = linalg.hadamard(g)
    encode = H[gind - 1, :]

    ftw = dinf(d1, d2) / tb  # fractional transition width of the slab profile

    if gind == 1:  # no sub-slices
        b = dzls(n, tb, d1, d2)
    else:
        # left stopband
        f = np.asarray([0, shift - (1 + ftw) * (tb / 2)])
        m = np.asarray([0, 0])
        w = np.asarray([d1 / d2])
        # first sub-band
        ii = 1
        gcent = shift + (ii - g / 2 - 1 / 2) * tb / g  # first band center
        # first band left edge
        f = np.append(f, gcent - (tb / g / 2 - ftw * (tb / 2)))
        m = np.append(m, encode[ii - 1])
        if encode[ii - 1] != encode[ii]:
            # add the first band's right edge and its amplitude, and a weight
            f = np.append(f, gcent + (tb / g / 2 - ftw * (tb / 2)))
            m = np.append(m, encode[ii - 1])
            w = np.append(w, 1)
        # middle sub-bands
        for ii in range(2, g):
            gcent = shift + (ii - g / 2 - 1 / 2) * tb / g  # center of band
            if encode[ii - 1] != encode[ii - 2]:
                # add a left edge and amp for this band
                f = np.append(f, gcent - (tb / g / 2 - ftw * (tb / 2)))
                m = np.append(m, encode[ii - 1])
            if encode[ii - 1] != encode[ii]:
                # add a right edge and its amp, and a weight for this band
                f = np.append(f, gcent + (tb / g / 2 - ftw * (tb / 2)))
                m = np.append(m, encode[ii - 1])
                w = np.append(w, 1)
        # last sub-band
        ii = g
        gcent = shift + (ii - g / 2 - 1 / 2) * tb / g  # center of last band
        if encode[ii - 1] != encode[ii - 2]:
            # add a left edge and amp for the last band
            f = np.append(f, gcent - (tb / g / 2 - ftw * (tb / 2)))
            m = np.append(m, encode[ii - 1])
        # add a right edge and its amp, and a weight for the last band
        f = np.append(f, gcent + (tb / g / 2 - ftw * (tb / 2)))
        m = np.append(m, encode[ii - 1])
        w = np.append(w, 1)
        # right stop-band
        f = np.append(f, (shift + (1 + ftw) * (tb / 2), (n / 2))) / (n / 2)
        m = np.append(m, [0, 0])
        w = np.append(w, d1 / d2)

        # separate the positive and negative bands
        mp = (m > 0).astype(float)
        mn = (m < 0).astype(float)

        # design the positive and negative filters
        c = np.exp(
            1j
            * 2
            * np.pi
            / (2 * (n + 1))
            * np.concatenate(
                [np.arange(0, n / 2 + 1, 1), np.arange(-n / 2, 0, 1)]
            )
        )
        bp = signal.firls(n + 1, f, mp, w)  # the positive filter
        bn = signal.firls(n + 1, f, mn, w)  # the negative filter

        # combine the filters and demodulate
        b = sp.ifft(
            np.multiply(sp.fft(bp - bn, center=False), c), center=False
        )

        b = np.real(b[:n])
        # hilbert transform to suppress negative passband
        b = signal.hilbert(b)
        # demodulate to DC
        c_shift = np.exp(-1j * 2 * np.pi / n * shift * np.arange(0, n, 1)) / 2
        c_shift *= np.exp(-1j * np.pi / n * shift)
        b = np.multiply(b, c_shift)

    return b


def dz_gslider_rf(
    n=256,
    g=5,
    flip=np.pi / 2,
    phi=np.pi,
    tb=12,
    d1=0.01,
    d2=0.01,
    cancel_alpha_phs=True,
):
    r"""Design a g-slider pulse rf

    Args:
        n (int): number of time points.
        g (int): number of sub-slices.
        flip (float): flip angle.
        phi (float): subslice phase.
        tb (int): time bandwidth product.
        d1 (float): passband ripple level in :math:'M_0^{-1}'.
        d2 (float): stopband ripple level in :math:'M_0^{-1}'.
        cancel_alpha_phs (bool): absorb the alpha phase
            profile from beta's profile, so they cancel for a flatter
            total phase

    Returns:
        rf (array): rf pulse out.

    References:
        Setsompop, K. et al. 'High-resolution in vivo diffusion imaging of the
        human brain with generalized slice dithered enhanced resolution:
        Simultaneous multislice (gSlider-SMS). Magn. Reson. Med.79, 141–151
        (2018).
    """
    bsf = np.sin(flip / 2)  # beta scaling factor

    rf = np.zeros((n, g), dtype="complex")
    for gind in range(1, g + 1):
        b = bsf * dz_gslider_b(n, g, gind, tb, d1, d2, phi)
        rf[:, gind - 1] = b2rf(b, cancel_alpha_phs)

    return rf


def b2rf(b, cancel_alpha_phs=False):
    a = b2a(b)
    if cancel_alpha_phs:
        b_a_phase = sp.fft(b, center=False, norm=None) * np.exp(
            -1j
            * np.angle(sp.fft(a[np.size(a) :: -1], center=False, norm=None))
        )
        b = sp.ifft(b_a_phase, center=False, norm=None)
    rf = ab2rf(a, b)

    return rf


def b2a(b):
    n = np.size(b)

    npad = n * 16
    bcp = np.zeros(npad, dtype=complex)
    bcp[0:n:1] = b
    bf = sp.fft(bcp, center=False, norm=None)
    bfmax = np.max(np.abs(bf))
    if bfmax >= 1:
        bf = bf / (1e-7 + bfmax)
    afa = mag2mp(np.sqrt(1 - np.abs(bf) ** 2))
    a = sp.fft(afa, center=False, norm=None) / npad
    a = a[0:n:1]
    a = a[::-1]

    return a


def mag2mp(x):
    n = np.size(x)
    xl = np.log(np.abs(x))  # Log of mag spectrum
    xlf = sp.fft(xl, center=False, norm=None)
    xlfp = xlf
    xlfp[0] = xlf[0]  # Keep DC the same
    xlfp[1 : (n // 2) : 1] = (
        2 * xlf[1 : (n // 2) : 1]
    )  # Double positive frequencies
    xlfp[n // 2] = xlf[n // 2]  # keep half Nyquist the same
    xlfp[n // 2 + 1 : n : 1] = 0  # zero negative frequencies
    xlaf = sp.ifft(xlfp, center=False, norm=None)
    a = np.exp(xlaf)  # complex exponentiation

    return a


def ab2rf(a, b):
    n = np.size(a)
    rf = np.zeros(n, dtype=complex)

    a = a.astype(complex)
    b = b.astype(complex)

    for ii in range(n - 1, -1, -1):
        cj = np.sqrt(1 / (1 + np.abs(b[ii] / a[ii]) ** 2))
        sj = np.conj(cj * b[ii] / a[ii])
        theta = np.arctan2(np.abs(sj), cj)
        psi = np.angle(sj)
        rf[ii] = 2 * theta * np.exp(1j * psi)

        # remove this rotation from polynomials
        if ii > 0:
            at = cj * a + sj * b
            bt = -np.conj(sj) * a + cj * b
            a = at[1 : ii + 1 : 1]
            b = bt[0:ii:1]

    return rf


def root_flip(b, d1, flip, tb, verbose=False):
    r"""Exhaustive root-flip pattern search for min-peak b1

    Args:
        b (array): SLR beta parameter.
        d1 (float): passband ripple level.
        flip (array): target flip angle.
        tb (int): pulse time bandwidth product.
        verbose (bool): print feeback on iterations.

    Returns:
        2-element tuple containing

        - **rf_out** (*array*): rf pulse out.
        - **b_out** (*array*): SLR beta parameter.

    References:
        Sharma, A. Lustig, M. and Grissom, W. (2016).
        Root-flipped multiband refocusing pulses.
        Magn Reson Med. 2016 Jan; 75(1): 227-237.
    """

    n = np.size(b)
    [_, b_resp] = signal.freqz(b)
    b /= np.max(np.abs(b_resp))  # normalize beta
    b *= np.sin(flip / 2 + np.arctan(d1 * 2) / 2)  # scale to target flip
    r = sp.util.leja(np.roots(b))

    candidates = np.logical_and(
        np.abs(1 - np.abs(r)) > 0.004, np.abs(np.angle(r)) < tb / n * np.pi
    )

    ii_min = 0
    ii_max = 2 ** np.sum(candidates)

    max_rf = np.max(np.abs(b2rf(b)))

    for ii in range(ii_min, ii_max):
        if ii % 20 == 0:
            if verbose:
                print(
                    "Evaluating root-flip pattern "
                    + str(ii)
                    + " out of "
                    + str(ii_max - ii_min)
                )

        # get a binary flipping pattern
        do_flip_str = format(ii, "b")
        do_flip = np.zeros(np.sum(candidates), dtype=bool)
        for jj in range(0, len(do_flip_str)):
            do_flip[jj] = bool(int(do_flip_str[jj]))

        # embed the pattern in an all-roots vector
        tmp = np.zeros(n - 1, dtype=bool)
        tmp[candidates] = do_flip
        do_flip = tmp

        # flip those indices
        r_flip = np.zeros(np.shape(r), dtype=complex)
        r_flip[:] = r[:]
        r_flip[do_flip] = np.conj(1 / r_flip[do_flip])

        b_tmp = np.poly(r_flip)
        [_, b_tmp_resp] = signal.freqz(b_tmp)
        b_tmp /= np.max(np.abs(b_tmp_resp))  # normalize beta
        b_tmp *= np.sin(flip / 2 + np.arctan(d1 * 2) / 2)  # scale to targ flip
        rf_tmp = b2rf(b_tmp)

        if np.max(np.abs(rf_tmp)) < max_rf:
            max_rf = np.max(np.abs(rf_tmp))
            rf_out = rf_tmp
            b_out = b_tmp

    return rf_out, b_out


def dz_recursive_rf(
    n_seg,
    tb,
    n,
    se_seq=False,
    tb_ref=8,
    z_pad_fact=4,
    win_fact=1.75,
    cancel_alpha_phs=True,
    t1=np.inf,
    tr_seg=60,
    use_mz=True,
    d1=0.01,
    d2=0.01,
    d1se=0.01,
    d2se=0.01,
):
    r"""Recursive SLR pulse design.

    Args:
        n_seg (int): number of segments designed by recursion.
        tb (int): time bandwidth product.
        n (int): pulse length.
        se_seq (bool): spin echo sequence.
        tb_ref (int): time bandwidth product of refocusing pulse.
        z_pad_fact (float): zero padding factor.
        win_fact (float): applied window factor.
        cancel_alpha_phs (bool): absorb the alpha phase
            profile from beta's profile, so they cancel for a flatter
            total phase
        t1 (float): t1
        tr_seg (int): length of tr
        use_mz (bool): design the pulses accounting for the actual Mz profile
        d1 (float): passband ripple level in :math:'M_0^{-1}'.
        d2 (float): stopband ripple level in :math:'M_0^{-1}'.
        d1se (float): passband ripple level for se
        d2se (float): stopband ripple level for se

    Returns:
        If se_seq=True, 2-element tuple containing

        - **rf** (*array*): rf pulse out.
        - **rf_ref** (*array*): rf refocusing pulse out.
    """

    # get refocusing pulse and its rotation parameters
    if se_seq is True:
        [bsf, d1se, d2se] = calc_ripples("se", d1se, d2se)
        b_ref = bsf * dzls(n, tb_ref, d1se, d2se)
        b_ref = np.concatenate(
            (
                np.zeros(int(z_pad_fact * n / 2 - n / 2)),
                b_ref,
                np.zeros(int(z_pad_fact * n / 2 - n / 2)),
            )
        )
        rf_ref = b2rf(b_ref)
        bref = sp.fft(b_ref, norm=None)
        bref /= np.max(np.abs(bref))
        bref_mag = np.abs(bref)
        aref_mag = np.abs(np.sqrt(1 - bref_mag**2))
        flip_ref = (
            2 * np.arcsin(bref_mag[int(z_pad_fact * n / 2)]) * 180 / np.pi
        )

    # get flip angles
    flip = np.zeros(n_seg)
    flip[n_seg - 1] = 90
    for jj in range(n_seg - 2, -1, -1):
        if se_seq is False:
            flip[jj] = np.arctan(np.sin(flip[jj + 1] * np.pi / 180))
            flip[jj] = flip[jj] * 180 / np.pi  # deg
        else:
            flip[jj] = np.arctan(
                np.cos(flip_ref * np.pi / 180)
                * np.sin(flip[jj + 1] * np.pi / 180)
            )
            flip[jj] = flip[jj] * 180 / np.pi  # deg

    # design first RF pulse
    b = np.zeros((int(z_pad_fact * n), n_seg), dtype=complex)
    b[
        int(z_pad_fact * n / 2 - n / 2) : int(z_pad_fact * n / 2 + n / 2), 0
    ] = dzls(n, tb, d1, d2)
    # b = np.concatenate((np.zeros(int(zPadFact*N/2-N/2)), b,
    #    np.zeros(int(zPadFact*N/2-N/2))))
    B = sp.fft(b[:, 0], norm=None)

    c = np.exp(
        -1j
        * 2
        * np.pi
        / (n * z_pad_fact)
        / 2
        * np.arange(-n * z_pad_fact / 2, n * z_pad_fact / 2, 1)
    )
    B = np.multiply(B, c)
    b[:, 0] = sp.ifft(B / np.max(np.abs(B)), norm=None)
    b[:, 0] *= np.sin(flip[0] * (np.pi / 180) / 2)
    rf = np.zeros((z_pad_fact * n, n_seg), dtype=complex)
    a = b2a(b[:, 0])
    if cancel_alpha_phs:
        # cancel a phase by absorbing into b
        # Note that this is the only time we need to do it
        b_a_phase = sp.fft(b[:, 0], center=False, norm=None) * np.exp(
            -1j
            * np.angle(sp.fft(a[np.size(a) :: -1], center=False, norm=None))
        )
        b[:, 0] = sp.ifft(b_a_phase, center=False, norm=None)
    rf[:, 0] = b2rf(b[:, 0])

    # get min-phase alpha and its response
    # a = b2a(b[:, 0])
    A = sp.fft(a)

    # calculate beta filter response
    B = sp.fft(b[:, 0], norm=None)

    if win_fact < z_pad_fact:
        win_len = (win_fact - 1) * n
        npad = n * z_pad_fact - win_fact * n
        # blackman window?
        window = signal.blackman(int((win_fact - 1) * n))
        # split in half; stick N ones in the middle
        window = np.concatenate(
            (
                window[0 : int(win_len / 2)],
                np.ones(n),
                window[int(win_len / 2) :],
            )
        )
        window = np.concatenate(
            (np.zeros(int(npad / 2)), window, np.zeros(int(npad / 2)))
        )
        # apply windowing to first pulse for consistency
        b[:, 0] = np.multiply(b[:, 0], window)
        rf[:, 0] = b2rf(b[:, 0])
        # recalculate B and A
        B = sp.fft(b[:, 0], norm=None)
        A = sp.fft(b2a(b[:, 0]), norm=None)

    # use A and B to get Mxy
    # Mxy = np.zeros((zPadFact*N, Nseg), dtype = complex)
    if se_seq is False:
        mxy0 = 2 * np.conj(A) * B
    else:
        mxy0 = 2 * A * np.conj(B) * bref**2

    # Amplitude of next pulse's Mxy profile will be
    #               |Mz*2*a*b| = |Mz*2*sqrt(1-abs(B).^2)*B|.
    # If we set this = |Mxy_1|, we can solve for |B| via solving quadratic
    # equation 4*Mz^2*(1-B^2)*B^2 = |Mxy_1|^2.
    # Subsequently solve for |A|, and get phase of A via min-phase, and
    # then get phase of B by dividing phase of A from first pulse's Mxy phase.
    mz = np.ones((z_pad_fact * n), dtype=complex)
    for jj in range(1, n_seg):
        # calculate Mz profile after previous pulse
        if se_seq is False:
            mz = mz * (1 - 2 * np.abs(B) ** 2) * np.exp(-tr_seg / t1) + (
                1 - np.exp(-tr_seg / t1)
            )
        else:
            mz = mz * (
                1 - 2 * (np.abs(A * bref_mag) ** 2 + np.abs(aref_mag * B) ** 2)
            )
            # (second term is about 1%)

        if use_mz is True:  # design the pulses accounting for the
            # actual Mz profile (the full method)
            # set up quadratic equation to get |B|
            cq = -np.abs(mxy0) ** 2
            if se_seq is False:
                bq = 4 * mz**2
                aq = -4 * mz**2
            else:
                bq = 4 * (bref_mag**4) * mz**2
                aq = -4 * (bref_mag**4) * mz**2
            bmag = np.sqrt(
                (-bq + np.real(np.sqrt(bq**2 - 4 * aq * cq))) / (2 * aq)
            )
            bmag[np.isnan(bmag)] = 0
            # get A - easier to get complex A than complex B since |A| is
            # determined by |B|, and phase is gotten by min-phase relationship
            # Phase of B doesn't matter here since only profile mag is used by
            # b2a
            A = sp.fft(b2a(sp.ifft(bmag, norm=None)), norm=None)
            # trick: now we can get complex B from ratio of Mxy and A
            B = mxy0 / (2 * np.conj(A) * mz)

        else:  # design assuming ideal Mz (conventional VFA)
            B *= np.sin(np.pi / 180 * flip[jj] / 2) / np.sin(
                np.pi / 180 * flip[jj - 1] / 2
            )
            A = sp.fft(b2a(sp.ifft(B, norm=None)), norm=None)

        # get polynomial
        b[:, jj] = sp.ifft(B, norm=None)

        if win_fact < z_pad_fact:
            b[:, jj] *= window
            # recalculate B and A
            B = sp.fft(b[:, jj], norm=None)
            A = sp.fft(b2a(b[:, jj]), norm=None)

        rf[:, jj] = b2rf(b[:, jj])

    # truncate the RF
    if win_fact < z_pad_fact:
        pulse_len = int(win_fact * n)
        rf = rf[int(npad / 2) : int(npad / 2 + pulse_len), :]

    if se_seq is False:
        return rf
    else:
        return rf, rf_ref



# -*- coding: utf-8 -*-
""":math:`B_1^{+}`-selective RF Pulse Design functions.

"""
import numpy as np

from sigpy.mri.rf import slr as slr
from sigpy.mri.rf.util import dinf

__all__ = ["dz_b1_rf", "dz_b1_gslider_rf", "dz_b1_hadamard_rf"]


def dz_b1_rf(
    dt=2e-6,
    tb=4,
    ptype="st",
    flip=np.pi / 6,
    pbw=0.3,
    pbc=2,
    d1=0.01,
    d2=0.01,
    os=8,
    split_and_reflect=True,
):
    """Design a :math:`B_1^{+}`-selective excitation pulse following Grissom \
    JMR 2014

    Args:
        dt (float): hardware sampling dwell time in s.
        tb (int): time-bandwidth product.
        ptype (string): pulse type, 'st' (small-tip excitation), 'ex' (pi/2
            excitation pulse), 'se' (spin-echo pulse), 'inv' (inversion), or
            'sat' (pi/2 saturation pulse).
        flip (float): flip angle, in radians.
        pbw (float): width of passband in Gauss.
        pbc (float): center of passband in Gauss.
        d1 (float): passband ripple level in :math:`M_0^{-1}`.
        d2 (float): stopband ripple level in :math:`M_0^{-1}`.
        os (int): matrix scaling factor.
        split_and_reflect (bool): option to split and reflect designed pulse.

    Split-and-reflect preserves pulse selectivity when scaled to excite large
    tip-angles.

    Returns:
        2-element tuple containing

        - **om1** (*array*): AM waveform.
        - **dom** (*array*): FM waveform (radians/s).

    References:
        Grissom, W., Cao, Z., & Does, M. (2014).
        :math:`B_1^{+}`-selective excitation pulse design using the Shinnar-Le
        Roux algorithm. Journal of Magnetic Resonance, 242, 189-196.
    """

    # calculate beta filter ripple
    [_, d1, d2] = slr.calc_ripples(ptype, d1, d2)

    # calculate pulse duration
    b = 4257 * pbw
    pulse_len = tb / b

    # calculate number of samples in pulse
    n = int(np.ceil(pulse_len / dt / 2) * 2)

    if pbc == 0:
        # we want passband as close to zero as possible.
        # do my own dual-band filter design to minimize interaction
        # between the left and right bands

        # build system matrix
        A = np.exp(
            1j
            * 2
            * np.pi
            * np.outer(
                np.arange(-n * os / 2, n * os / 2), np.arange(-n / 2, n / 2)
            )
            / (n * os)
        )

        # build target pattern
        ii = np.arange(-n * os / 2, n * os / 2) / (n * os) * 2
        w = dinf(d1, d2) / tb
        f = np.asarray([0, (1 - w) * (tb / 2), (1 + w) * (tb / 2), n / 2]) / (
            n / 2
        )
        d = np.double(np.abs(ii) < f[1])
        ds = np.double(np.abs(ii) > f[2])

        # shift the target pattern to minimum center position
        pbc = int(np.ceil((f[2] - f[1]) * n * os / 2 + f[1] * n * os / 2))
        dl = np.roll(d, pbc)
        dr = np.roll(d, -pbc)
        dsl = np.roll(ds, pbc)
        dsr = np.roll(ds, -pbc)

        # build error weight vector
        w = dl + dr + d1 / d2 * np.multiply(dsl, dsr)

        # solve for the dual-band filter
        AtA = A.conj().T @ np.multiply(np.reshape(w, (np.size(w), 1)), A)
        Atd = A.conj().T @ np.multiply(w, dr - dl)
        h = np.imag(np.linalg.pinv(AtA) @ Atd)

    else:  # normal design
        # design filter
        h = slr.dzls(n, tb, d1, d2)

        # dual-band-modulate the filter
        om = 2 * np.pi * 4257 * pbc  # modulation frequency
        t = np.arange(0, n) * pulse_len / n - pulse_len / 2
        h = 2 * h * np.sin(om * t)

    if split_and_reflect:
        # split and flip fm waveform to improve large-tip accuracy
        dom = np.concatenate((h[n // 2 :: -1], h, h[n : n // 2 : -1])) / 2
    else:
        dom = np.concatenate((0 * h[n // 2 :: -1], h, 0 * h[n : n // 2 : -1]))

    # scale to target flip, convert to Hz
    dom = dom * flip / (2 * np.pi * dt)

    # build am waveform
    om1 = np.concatenate((-np.ones(n // 2), np.ones(n), -np.ones(n // 2)))

    return om1, dom


def dz_b1_gslider_rf(
    dt=2e-6,
    g=5,
    tb=12,
    ptype="st",
    flip=np.pi / 6,
    pbw=0.5,
    pbc=2,
    d1=0.01,
    d2=0.01,
    split_and_reflect=True,
):
    """Design a :math:`B_1^{+}`-selective excitation gSlider pulse following
     Grissom JMR 2014.

    Args:
        dt (float): hardware sampling dwell time in s.
        g (int): number of slabs to be acquired.
        tb (int): time-bandwidth product.
        ptype (string): pulse type, 'st' (small-tip excitation), 'ex' (pi/2
            excitation pulse), 'se' (spin-echo pulse), 'inv' (inversion), or
            'sat' (pi/2 saturation pulse).
        flip (float): flip angle, in radians.
        pbw (float): width of passband in Gauss.
        pbc (float): center of passband in Gauss.
        d1 (float): passband ripple level in :math:`M_0^{-1}`.
        d2 (float): stopband ripple level in :math:`M_0^{-1}`.
        split_and_reflect (bool): option to split and reflect designed pulse.

    Split-and-reflect preserves pulse selectivity when scaled to excite large
     tip-angles.

    Returns:
        2-element tuple containing

        - **om1** (*array*): AM waveform.
        - **dom** (*array*): FM waveform (radians/s).

    References:
        Grissom, W., Cao, Z., & Does, M. (2014).
        :math:`B_1^{+}`-selective excitation pulse design using the Shinnar-Le
        Roux algorithm. Journal of Magnetic Resonance, 242, 189-196.
    """

    # calculate beta filter ripple
    [_, d1, d2] = slr.calc_ripples(ptype, d1, d2)
    # if ptype == 'st':
    bsf = flip

    # calculate pulse duration
    b = 4257 * pbw
    pulse_len = tb / b

    # calculate number of samples in pulse
    n = int(np.ceil(pulse_len / dt / 2) * 2)

    om = 2 * np.pi * 4257 * pbc  # modulation freq to center profile at pbc
    t = np.arange(0, n) * pulse_len / n - pulse_len / 2

    om1 = np.zeros((2 * n, g))
    dom = np.zeros((2 * n, g))
    for gind in range(1, g + 1):
        # design filter
        h = bsf * slr.dz_gslider_b(n, g, gind, tb, d1, d2, np.pi, n // 4)

        # modulate filter to center and add it to a time-reversed and modulated
        # copy, then take the imaginary part to get an odd filter
        h = np.imag(h * np.exp(1j * om * t) - h[n::-1] * np.exp(1j * -om * t))
        if split_and_reflect:
            # split and flip fm waveform to improve large-tip accuracy
            dom[:, gind - 1] = (
                np.concatenate((h[n // 2 :: -1], h, h[n : n // 2 : -1])) / 2
            )
        else:
            dom[:, gind - 1] = np.concatenate(
                (0 * h[n // 2 :: -1], h, 0 * h[n : n // 2 : -1])
            )
        # build am waveform
        om1[:, gind - 1] = np.concatenate(
            (-np.ones(n // 2), np.ones(n), -np.ones(n // 2))
        )

    # scale to target flip, convert to Hz
    dom = dom / (2 * np.pi * dt)

    return om1, dom


def dz_b1_hadamard_rf(
    dt=2e-6,
    g=8,
    tb=16,
    ptype="st",
    flip=np.pi / 6,
    pbw=2,
    pbc=2,
    d1=0.01,
    d2=0.01,
    split_and_reflect=True,
):
    """Design a :math:`B_1^{+}`-selective Hadamard-encoded pulse following \
     Grissom JMR 2014.
    Args:
        dt (float): hardware sampling dwell time in s.
        g (int): number of slabs to be acquired.
        tb (int): time-bandwidth product.
        ptype (string): pulse type, 'st' (small-tip excitation), 'ex' (pi/2 \
            excitation pulse), 'se' (spin-echo pulse), 'inv' (inversion), or \
            'sat' (pi/2 saturation pulse).
        flip (float): flip angle, in radians.
        pbw (float): width of passband in Gauss.
        pbc (float): center of passband in Gauss.
        d1 (float): passband ripple level in :math:`M_0^{-1}`.
        d2 (float): stopband ripple level in :math:`M_0^{-1}`.
        split_and_reflect (bool): option to split and reflect designed pulse.

    Split-and-reflect preserves pulse selectivity when scaled to excite large
    tip-angles.

    Returns:
        2-element tuple containing

        - **om1** (*array*): AM waveform.
        - **dom** (*array*): FM waveform (radians/s).

    References:
        Grissom, W., Cao, Z., & Does, M. (2014).
        :math:`B_1^{+}`-selective excitation pulse design using the Shinnar-Le
        Roux algorithm. Journal of Magnetic Resonance, 242, 189-196.
    """

    # calculate beta filter ripple
    [_, d1, d2] = slr.calc_ripples(ptype, d1, d2)
    bsf = flip

    # calculate pulse duration
    b = 4257 * pbw
    pulse_len = tb / b

    # calculate number of samples in pulse
    n = int(np.ceil(pulse_len / dt / 2) * 2)

    # modulation frequency to center profile at pbc gauss
    om = 2 * np.pi * 4257 * pbc
    t = np.arange(0, n) * pulse_len / n - pulse_len / 2

    om1 = np.zeros((2 * n, g))
    dom = np.zeros((2 * n, g))
    for gind in range(1, g + 1):
        # design filter
        h = bsf * slr.dz_hadamard_b(n, g, gind, tb, d1, d2, n // 4)

        # modulate filter to center and add it to a time-reversed and modulated
        # copy, then take the imaginary part to get an odd filter
        h = np.imag(h * np.exp(1j * om * t) - h[n::-1] * np.exp(1j * -om * t))
        if split_and_reflect:
            # split and flip fm waveform to improve large-tip accuracy
            dom[:, gind - 1] = (
                np.concatenate((h[n // 2 :: -1], h, h[n : n // 2 : -1])) / 2
            )
        else:
            dom[:, gind - 1] = np.concatenate(
                (0 * h[n // 2 :: -1], h, 0 * h[n : n // 2 : -1])
            )
        # build am waveform
        om1[:, gind - 1] = np.concatenate(
            (-np.ones(n // 2), np.ones(n), -np.ones(n // 2))
        )

    # scale to target flip, convert to Hz
    dom = dom / (2 * np.pi * dt)

    return om1, dom



# -*- coding: utf-8 -*-
"""RF Pulse Simulation Functions.

"""
from sigpy import backend

__all__ = ["abrm", "abrm_nd", "abrm_hp", "abrm_ptx"]


def abrm(rf, x, balanced=False):
    r"""1D RF pulse simulation, with simultaneous RF + gradient rotations.

    Args:
         rf (array): rf waveform input.
         x (array): spatial locations.
         balanced (bool): toggles application of rewinder.

    Returns:
        2-element tuple containing

        - **a** (*array*): SLR alpha parameter.
        - **b** (*array*): SLR beta parameter.

    References:
        Pauly, J., Le Roux, Patrick., Nishimura, D., and Macovski, A.(1991).
        'Parameter Relations for the Shinnar-LeRoux Selective Excitation
        Pulse Design Algorithm'.
        IEEE Transactions on Medical Imaging, Vol 10, No 1, 53-65.
    """

    device = backend.get_device(rf)
    xp = device.xp
    with device:
        eps = 1e-16

        g = xp.ones(xp.size(rf)) * 2 * xp.pi / xp.size(rf)

        a = xp.ones(xp.size(x), dtype=complex)
        b = xp.zeros(xp.size(x), dtype=complex)
        for mm in range(xp.size(rf)):
            om = x * g[mm]
            phi = xp.sqrt(xp.abs(rf[mm]) ** 2 + om**2) + eps
            n = xp.column_stack(
                (xp.real(rf[mm]) / phi, xp.imag(rf[mm]) / phi, om / phi)
            )
            av = xp.cos(phi / 2) - 1j * n[:, 2] * xp.sin(phi / 2)
            bv = -1j * (n[:, 0] + 1j * n[:, 1]) * xp.sin(phi / 2)
            at = av * a - xp.conj(bv) * b
            bt = bv * a + xp.conj(av) * b
            a = at
            b = bt

        if balanced:  # apply a rewinder
            g = -2 * xp.pi / 2
            om = x * g
            phi = xp.abs(om) + eps
            nz = om / phi
            av = xp.cos(phi / 2) - 1j * nz * xp.sin(phi / 2)
            a = av * a
            b = xp.conj(av) * b

        return a, b


def abrm_nd(rf, x, g):
    r"""N-dim RF pulse simulation

    Assumes that x has inverse spatial units of g, and g has gamma*dt applied.

    Assumes dimensions x = [...,Ndim], g = [Ndim,Nt].

    Args:
         rf (array): rf waveform input.
         x (array): spatial locations.
         g (array): gradient array.

    Returns:
        2-element tuple containing

        - **a** (*array*): SLR alpha parameter.
        - **b** (*array*): SLR beta parameter.

    References:
        Pauly, J., Le Roux, Patrick., Nishimura, D., and Macovski, A.(1991).
        'Parameter Relations for the Shinnar-LeRoux Selective Excitation
        Pulse Design Algorithm'.
        IEEE Transactions on Medical Imaging, Vol 10, No 1, 53-65.
    """

    device = backend.get_device(rf)
    xp = device.xp
    with device:
        eps = 1e-16

        a = xp.ones(xp.shape(x)[0], dtype=complex)
        b = xp.zeros(xp.shape(x)[0], dtype=complex)
        for mm in range(xp.size(rf)):
            om = x @ g[mm, :]
            phi = xp.sqrt(xp.abs(rf[mm]) ** 2 + om**2)
            n = xp.column_stack(
                (
                    xp.real(rf[mm]) / (phi + eps),
                    xp.imag(rf[mm]) / (phi + eps),
                    om / (phi + eps),
                )
            )
            av = xp.cos(phi / 2) - 1j * n[:, 2] * xp.sin(phi / 2)
            bv = -1j * (n[:, 0] + 1j * n[:, 1]) * xp.sin(phi / 2)
            at = av * a - xp.conj(bv) * b
            bt = bv * a + xp.conj(av) * b
            a = at
            b = bt

        return a, b


def abrm_hp(rf, gamgdt, xx, dom0dt=0):
    r"""1D RF pulse simulation, with non-simultaneous RF + gradient rotations.

    Args:
        rf (array): rf pulse samples in radians.
        gamdt (array): gradient samples in radians/(units of xx).
        xx (array): spatial locations.
        dom0dt (float): off-resonance phase in radians.

    Returns:
        2-element tuple containing

        - **a** (*array*): SLR alpha parameter.
        - **b** (*array*): SLR beta parameter.

    References:
        Pauly, J., Le Roux, Patrick., Nishimura, D., and Macovski, A.(1991).
        'Parameter Relations for the Shinnar-LeRoux Selective Excitation
        Pulse Design Algorithm'.
        IEEE Transactions on Medical Imaging, Vol 10, No 1, 53-65.
    """

    device = backend.get_device(rf)
    xp = device.xp
    with device:
        Ns = xp.shape(xx)
        Ns = Ns[0]  # Ns: # of spatial locs
        Nt = xp.shape(gamgdt)
        Nt = Nt[0]  # Nt: # time points

        a = xp.ones((Ns,))
        b = xp.zeros((Ns,))

        for ii in xp.arange(Nt):
            # apply phase accural
            z = xp.exp(-1j * (xx * gamgdt[ii,] + dom0dt))
            b = b * z

            # apply rf
            C = xp.cos(xp.abs(rf[ii]) / 2)
            S = 1j * xp.exp(1j * xp.angle(rf[ii])) * xp.sin(xp.abs(rf[ii]) / 2)
            at = a * C - b * xp.conj(S)
            bt = a * S + b * C

            a = at
            b = bt

        z = xp.exp(1j / 2 * (xx * xp.sum(gamgdt, axis=0) + Nt * dom0dt))
        a = a * z
        b = b * z

        return a, b


def abrm_ptx(b1, x, g, dt, fmap=None, sens=None):
    r"""N-dim RF pulse simulation

    Assumes that x has inverse spatial units of g, and g has gamma*dt applied.

    Assumes dimensions rf = [Nc, Nt], x = [...,Ndim], g = [Ndim,Nt], and
    sens = [Nc, dim, dim].

    Args:
         b1 (array): rf waveform input samples in radians.
         x (array): spatial locations (m).
         g (array): gradient array (mT/m with gamma*dt applied).
         dt (float): hardware dwell time (s).
         fmap (array): off-resonance map (Hz).
         sens (array or None): B1+ sensitivity matrix. If None, creates matrix
            of 1's. Input size [Nc dim dim]


    Returns:
        4-element tuple containing

        - **a** (*array*): SLR alpha parameter.
        - **b** (*array*): SLR beta parameter.
        - **m** (*array*): transverse magnetization.
        - **mz** (*array*): longitudinal magnetization.

    References:
        Pauly, J., Le Roux, Patrick., Nishimura, D., and Macovski, A.(1991).
        'Parameter Relations for the Shinnar-LeRoux Selective Excitation
        Pulse Design Algorithm'.
        IEEE Transactions on Medical Imaging, Vol 10, No 1, 53-65.

        Grissom, W., Xu, D., Kerr, A., Fessler, J. and Noll, D. (2009). 'Fast
        large-tip-angle multidimensional and parallel RF pulse design in MRI'
        IEEE Trans Med Imaging, Vol 28, No 10, 1548-59.
    """

    device = backend.get_device(b1)
    xp = device.xp
    with device:
        gam = 267.522 * 1e6 / 1000  # rad/s/mT

        dim = int(xp.sqrt(x.shape[0]))
        Ns = dim * dim
        Nc = b1.shape[0]
        Nt = b1.shape[1]
        dim = int(xp.sqrt(x.shape[0]))

        if sens is None:
            sens = xp.ones((dim * dim, Nc))
        else:
            sens = xp.transpose(sens)
            sens = xp.reshape(sens, (dim * dim, Nc))

        bxy = sens @ b1
        bz = x @ xp.transpose(g)

        if fmap is not None and xp.sum(xp.abs(fmap)) != 0:
            rep_b0 = xp.repeat(xp.expand_dims(fmap.flatten(), 0), Nt, axis=0)
            bz += xp.transpose(rep_b0 / gam * 2 * xp.pi)

        statea = xp.ones((Ns, 1))
        stateb = xp.zeros((Ns, 1))
        a = xp.ones(xp.shape(x)[0], dtype=complex)
        b = xp.zeros(xp.shape(x)[0], dtype=complex)
        for mm in range(Nt):
            phi = dt * gam * xp.sqrt(xp.abs(bxy[:, mm]) ** 2 + bz[:, mm] ** 2)
            with xp.errstate(divide="ignore"):
                normfact = dt * gam * (phi**-1)
                normfact[xp.isinf(normfact)] = 0
                nxy = normfact * bxy[:, mm]
                nxy[xp.isinf(nxy)] = 0
            nz = normfact * bz[:, mm]
            nz[xp.isinf(nz)] = 0
            cp = xp.cos(phi / 2)
            sp = xp.sin(phi / 2)
            alpha = xp.expand_dims(cp + 1j * nz * sp, 1)
            beta = xp.expand_dims(1j * xp.conj(nxy) * sp, 1)

            tmpa = xp.multiply(alpha, statea) + xp.multiply(beta, stateb)
            tmpb = -xp.conj(beta) * statea + xp.conj(alpha) * stateb

            statea, stateb = tmpa, tmpb

            # NOT returning all states:
            a = statea
            b = -xp.conj(stateb)

        mxy0 = 0 + 1j * 0
        mz0 = 1
        m = mz0 * xp.conj(statea) * stateb
        m += mxy0 * xp.conj(statea) ** 2
        m -= xp.conj(mxy0) * (stateb**2)
        mz = mz0 * (statea * xp.conj(statea) - stateb * xp.conj(stateb))
        mz += 2 * xp.real(
            mxy0 * xp.conj(statea) * xp.negative(xp.conj(stateb))
        )

        return a, b, m, mz



# -*- coding: utf-8 -*-
"""MRI pulse-design-specific linear operators.
"""
import sigpy as sp
from sigpy import backend


def PtxSpatialExplicit(sens, coord, dt, img_shape, b0=None, ret_array=False):
    """Explicit spatial-domain pulse design linear operator.
    Linear operator relates rf pulses to desired magnetization.
    Equivalent matrix has dimensions [Ns Nt].

    Args:
        sens (array): sensitivity maps. [nc dim dim]
        coord (None or array): coordinates. [nt 2]
        dt (float): hardware sampling dt.
        img_shape (None or tuple): image shape.
        b0 (array): 2D array, B0 inhomogeneity map.
        ret_array (bool): if true, return explicit numpy array.
            Else return linop.

    Returns:
        SigPy linop with A.repr_string 'pTx spatial explicit', or numpy array
        if selected with 'ret_array'


    References:
        Grissom, W., Yip, C., Zhang, Z., Stenger, V. A., Fessler, J. A.
        & Noll, D. C.(2006).
        Spatial Domain Method for the Design of RF Pulses in Multicoil
        Parallel Excitation. Magnetic resonance in medicine, 56, 620-629.
    """
    three_d = False
    if len(img_shape) >= 3:
        three_d = True

    device = backend.get_device(sens)
    xp = device.xp
    with device:
        nc = sens.shape[0]
        dur = dt * coord.shape[0]  # duration of pulse, in s

        # create time vector
        t = xp.expand_dims(xp.linspace(0, dur, coord.shape[0]), axis=1)

        # row-major order
        # x L to R, y T to B
        x_ = xp.linspace(
            -img_shape[0] / 2, img_shape[0] - img_shape[0] / 2, img_shape[0]
        )
        y_ = xp.linspace(
            img_shape[1] / 2, -(img_shape[1] - img_shape[1] / 2), img_shape[1]
        )
        if three_d:
            z_ = xp.linspace(
                -img_shape[2] / 2,
                img_shape[2] - img_shape[2] / 2,
                img_shape[2],
            )
            x, y, z = xp.meshgrid(x_, y_, z_, indexing="ij")
        else:
            x, y = xp.meshgrid(x_, y_, indexing="ij")

        # create explicit Ns * Nt system matrix, for 3d or 2d problem
        if three_d:
            if b0 is None:
                AExplicit = xp.exp(
                    1j
                    * (
                        xp.outer(x.flatten(), coord[:, 0])
                        + xp.outer(y.flatten(), coord[:, 1])
                        + xp.outer(z.flatten(), coord[:, 2])
                    )
                )
            else:
                AExplicit = xp.exp(
                    1j * 2 * xp.pi * xp.transpose(b0.flatten() * (t - dur))
                    + 1j
                    * (
                        xp.outer(x.flatten(), coord[:, 0])
                        + xp.outer(y.flatten(), coord[:, 1])
                        + xp.outer(z.flatten(), coord[:, 2])
                    )
                )
        else:
            if b0 is None:
                AExplicit = xp.exp(
                    1j
                    * (
                        xp.outer(x.flatten(), coord[:, 0])
                        + xp.outer(y.flatten(), coord[:, 1])
                    )
                )
            else:
                AExplicit = xp.exp(
                    1j * 2 * xp.pi * xp.transpose(b0.flatten() * (t - dur))
                    + 1j
                    * (
                        xp.outer(x.flatten(), coord[:, 0])
                        + xp.outer(y.flatten(), coord[:, 1])
                    )
                )

        # add sensitivities to system matrix
        AFullExplicit = xp.empty(AExplicit.shape)
        for ii in range(nc):
            if three_d:
                tmp = xp.squeeze(sens[ii, :, :, :]).flatten()
            else:
                tmp = sens[ii, :, :].flatten()
            D = xp.transpose(xp.tile(tmp, [coord.shape[0], 1]))
            AFullExplicit = xp.concatenate(
                (AFullExplicit, D * AExplicit), axis=1
            )

        # remove 1st empty AExplicit entries
        AFullExplicit = AFullExplicit[:, coord.shape[0] :]
        A = sp.linop.MatMul((coord.shape[0] * nc, 1), AFullExplicit)

        # Finally, adjustment of input/output dimensions to be consistent with
        # the existing Sense linop operator. [nc x nt] in, [dim x dim] out
        Ro = sp.linop.Reshape(ishape=A.oshape, oshape=sens.shape[1:])
        Ri = sp.linop.Reshape(
            ishape=(nc, coord.shape[0]), oshape=(coord.shape[0] * nc, 1)
        )
        A = Ro * A * Ri

        A.repr_str = "pTx spatial explicit"

        # output a sigpy linop or a numpy array
        if ret_array:
            return A.linops[1].mat
        else:
            return A



# -*- coding: utf-8 -*-
"""MRI RF shimming.
"""

import numpy as np

import sigpy as sp
from sigpy import backend
from sigpy.mri import rf as rf

__all__ = ["calc_shims", "init_optimal_spectral", "init_circ_polar"]


def calc_shims(shim_roi, sens, x0, dt, lamb=0, max_iter=50):
    """RF shim designer. Uses the Gerchberg Saxton algorithm.

    Args:
       shim_roi (array): region within volume to be shimmed. Mask of 1's and
           0's. [dim_x dim_y dim_z]
       sens (array): sensitivity maps. [Nc dim_x dim_y dim_z]
       x0 (array) initial guess for shim values. [Nc 1]
       dt (float): hardware sampling dwell time.
       lamb (float): regularization term.
       max_iter (int): max number of iterations.

    Returns:
        Vector of complex shim weights.
    """

    k1 = np.expand_dims(np.array((0, 0, 0)), 0)
    A = rf.PtxSpatialExplicit(
        sens, coord=k1, dt=dt, img_shape=shim_roi.shape, ret_array=False
    )

    alg_method = sp.alg.GerchbergSaxton(
        A, shim_roi, x0, max_iter=max_iter, tol=10e-9, lamb=lamb
    )
    while not alg_method.done():
        alg_method.update()

    return alg_method.x


def init_optimal_spectral(A, sens, preproc=False):
    """Function to return initial shim weights based on an optimal spectral
    method, an eigenvector-based method.

        Args:
            A (linop): sigpy Linear operator.
            sens (array): sensitivity maps. [Nc dim_x dim_y]
            preproc (bool): option to apply preprocessing function before \
                finding eigenvectors

        Returns:
            Vector of complex shim weights.

        References:
            Chandra, R., Zhong, Z., Hontz, J., McCulloch, V., Studer, C.,
            Goldstein, T. (2017) 'PhasePack: A Phase Retrieval Library.'
            arXiv:1711.10175.
    """
    device = backend.get_device(sens)
    xp = device.xp
    with device:
        if hasattr(A, "repr_str") and A.repr_str == "pTx spatial explicit":
            Anum = A.linops[1].mat
        else:
            Anum = A

        sens = sens.flatten()
        n = Anum.shape[1]
        Anumt = xp.transpose(Anum)

        m = sens.size
        y = sens**2

        # normalize the measurements
        delta = m / n
        ymean = y / xp.mean(y)

        # apply pre-processing function
        yplus = xp.amax(y)
        Y = (1 / m) * Anumt @ Anum

        if preproc:
            T = (yplus - 1) / (yplus + xp.sqrt(delta) - 1)

            # unnormalize
            T *= ymean
            T = xp.transpose(xp.expand_dims(T, axis=1))

            for mm in range(m):
                col = Anum[mm, :]
                aat = col * xp.transpose(col)
                Y = Y + (1 / m) * T[mm] * aat

        w, v = xp.linalg.eigh(Y)

        return xp.expand_dims(v[:, 0], 1)


def init_circ_polar(sens):
    """Function to return circularly polarized initial shim weights. Provides
    shim weights that set the phase to be even in the middle of the sens
    profiles.

       Args:
           sens (array): sensitivity maps. [Nc dim_x dim_y]

       Returns:
           Vector of complex shim weights.
    """
    dim = sens.shape[1]
    device = backend.get_device(sens)
    xp = device.xp
    with device:
        # As a rough approximation, assume that the center of sens profile is
        # also the center of the object within the profile to be imaged.
        phs = xp.angle(sens[:, xp.int(dim / 2), xp.int(dim / 2)])
        phs_wt = xp.exp(-phs * 1j)

    return xp.expand_dims(phs_wt, 1)



import unittest

import numpy as np
import numpy.testing as npt

from sigpy import config, thresh

if config.cupy_enabled:
    import cupy as cp

if __name__ == "__main__":
    unittest.main()


class TestThresh(unittest.TestCase):
    def test_l2_proj(self):
        x = np.ones(5)
        y = np.full(5, 1 / 5**0.5)
        npt.assert_allclose(thresh.l2_proj(1, x), y)

        x = np.ones(5)
        y = np.ones(5)
        npt.assert_allclose(thresh.l2_proj(5**0.5, x), y)

        x = np.ones(5)
        y = np.ones(5)
        npt.assert_allclose(thresh.l2_proj(10, x), y)

    def test_linf_proj(self):
        x = np.ones(5)
        y = np.ones(5)
        npt.assert_allclose(thresh.linf_proj(1.1, x), y)

        x = np.ones(5)
        y = np.ones(5) * 0.1
        npt.assert_allclose(thresh.linf_proj(0.1, x), y)

        x = [-2, -1, 0, 1, 2]
        y = [-1, -1, 0, 1, 1]
        npt.assert_allclose(thresh.linf_proj(1, x), y)

    def test_psd_proj(self):
        x = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -2]])
        y = np.array([[1, 0, 0], [0, 0, 0], [0, 0, 0]])
        npt.assert_allclose(thresh.psd_proj(x), y)

    def test_soft_thresh(self):
        x = np.array([-2, -1.5, -1, 0.5, 0, 0.5, 1, 1.5, 2])
        y = np.array([-1, -0.5, 0, 0, 0, 0, 0, 0.5, 1])

        npt.assert_allclose(thresh.soft_thresh(1, x), y)

    def test_hard_thresh(self):
        x = np.array([-2, -1.5, -1, 0.5, 0, 0.5, 1, 1.5, 2])
        y = np.array([-2, -1.5, 0, 0, 0, 0, 0, 1.5, 2])

        npt.assert_allclose(thresh.hard_thresh(1, x), y)

    if config.cupy_enabled:

        def test_soft_thresh_cuda(self):
            x = cp.array([-2, -1.5, -1, 0.5, 0, 0.5, 1, 1.5, 2])
            y = cp.array([-1, -0.5, 0, 0, 0, 0, 0, 0.5, 1])
            lamda = cp.array([1.0])

            cp.testing.assert_allclose(thresh.soft_thresh(lamda, x), y)

        def test_hard_thresh_cuda(self):
            x = cp.array([-2, -1.5, -1, 0.5, 0, 0.5, 1, 1.5, 2])
            y = cp.array([-2, -1.5, 0, 0, 0, 0, 0, 1.5, 2])
            lamda = cp.array([1.0])

            cp.testing.assert_allclose(thresh.hard_thresh(lamda, x), y)
